import json
import os
import logging
import pycurl
import socket

from twisted.internet.defer import succeed, fail

from landscape.broker.registration import (
    InvalidCredentialsError, RegistrationHandler, is_cloud_managed, EC2_HOST,
    EC2_API, Identity)

from landscape.broker.config import BrokerConfiguration
from landscape.tests.helpers import LandscapeTest
from landscape.broker.tests.helpers import (
    BrokerConfigurationHelper, RegistrationHelper)
from landscape.lib.bpickle import dumps
from landscape.lib.fetch import HTTPCodeError, FetchError
from landscape.lib.persist import Persist
from landscape.lib.vm_info import get_vm_info
from landscape.configuration import print_text


class IdentityTest(LandscapeTest):

    helpers = [BrokerConfigurationHelper]

    def setUp(self):
        super(IdentityTest, self).setUp()
        self.persist = Persist(filename=self.makePersistFile())
        self.identity = Identity(self.config, self.persist)

    def check_persist_property(self, attr, persist_name):
        value = "VALUE"
        self.assertEqual(getattr(self.identity, attr), None,
                         "%r attribute should default to None, not %r" %
                         (attr, getattr(self.identity, attr)))
        setattr(self.identity, attr, value)
        self.assertEqual(getattr(self.identity, attr), value,
                         "%r attribute should be %r, not %r" %
                         (attr, value, getattr(self.identity, attr)))
        self.assertEqual(
            self.persist.get(persist_name), value,
            "%r not set to %r in persist" % (persist_name, value))

    def check_config_property(self, attr):
        value = "VALUE"
        setattr(self.config, attr, value)
        self.assertEqual(getattr(self.identity, attr), value,
                         "%r attribute should be %r, not %r" %
                         (attr, value, getattr(self.identity, attr)))

    def test_secure_id(self):
        self.check_persist_property("secure_id",
                                    "registration.secure-id")

    def test_insecure_id(self):
        self.check_persist_property("insecure_id",
                                    "registration.insecure-id")

    def test_computer_title(self):
        self.check_config_property("computer_title")

    def test_account_name(self):
        self.check_config_property("account_name")

    def test_registration_key(self):
        self.check_config_property("registration_key")

    def test_client_tags(self):
        self.check_config_property("tags")

    def test_access_group(self):
        self.check_config_property("access_group")


class RegistrationHandlerTestBase(LandscapeTest):

    helpers = [RegistrationHelper]

    def setUp(self):
        super(RegistrationHandlerTestBase, self).setUp()
        logging.getLogger().setLevel(logging.INFO)
        self.hostname = "ooga.local"
        self.addCleanup(setattr, socket, "getfqdn", socket.getfqdn)
        socket.getfqdn = lambda: self.hostname


class RegistrationHandlerTest(RegistrationHandlerTestBase):

    def test_server_initiated_id_changing(self):
        """
        The server must be able to ask a client to change its secure
        and insecure ids even if no requests were sent.
        """
        self.exchanger.handle_message(
            {"type": "set-id", "id": "abc", "insecure-id": "def"})
        self.assertEqual(self.identity.secure_id, "abc")
        self.assertEqual(self.identity.insecure_id, "def")

    def test_registration_done_event(self):
        """
        When new ids are received from the server, a "registration-done"
        event is fired.
        """
        reactor_mock = self.mocker.patch(self.reactor)
        reactor_mock.fire("registration-done")
        self.mocker.replay()
        self.exchanger.handle_message(
            {"type": "set-id", "id": "abc", "insecure-id": "def"})

    def test_unknown_id(self):
        self.identity.secure_id = "old_id"
        self.identity.insecure_id = "old_id"
        self.mstore.set_accepted_types(["register"])
        self.exchanger.handle_message({"type": "unknown-id"})
        self.assertEqual(self.identity.secure_id, None)
        self.assertEqual(self.identity.insecure_id, None)

    def test_unknown_id_with_clone(self):
        """
        If the server reports us that we are a clone of another computer, then
        set our computer's title accordingly.
        """
        self.config.computer_title = "Wu"
        self.mstore.set_accepted_types(["register"])
        self.exchanger.handle_message({"type": "unknown-id", "clone-of": "Wu"})
        self.assertEqual("Wu (clone)", self.config.computer_title)
        self.assertIn("Client is clone of computer Wu",
                      self.logfile.getvalue())

    def test_should_register(self):
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.assertTrue(self.handler.should_register())

    def test_should_register_with_existing_id(self):
        self.mstore.set_accepted_types(["register"])
        self.identity.secure_id = "secure"
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.assertFalse(self.handler.should_register())

    def test_should_register_without_computer_title(self):
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = None
        self.assertFalse(self.handler.should_register())

    def test_should_register_without_account_name(self):
        self.mstore.set_accepted_types(["register"])
        self.config.account_name = None
        self.assertFalse(self.handler.should_register())

    def test_should_register_with_unaccepted_message(self):
        self.assertFalse(self.handler.should_register())

    def test_queue_message_on_exchange(self):
        """
        When a computer_title and account_name are available, no
        secure_id is set, and an exchange is about to happen,
        queue a registration message.
        """
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        self.assertEqual(1, len(messages))
        self.assertEqual("register", messages[0]["type"])
        self.assertEqual(self.logfile.getvalue().strip(),
                         "INFO: Queueing message to register with account "
                         "'account_name' without a password.")

    def test_queue_message_on_exchange_with_vm_info(self):
        """
        When a computer_title and account_name are available, no
        secure_id is set, and an exchange is about to happen,
        queue a registration message with VM information.
        """
        get_vm_info_mock = self.mocker.replace(get_vm_info)
        get_vm_info_mock()
        self.mocker.result("vmware")
        self.mocker.replay()
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        self.assertEqual("vmware", messages[0]["vm-info"])
        self.assertEqual(self.logfile.getvalue().strip(),
                         "INFO: Queueing message to register with account "
                         "'account_name' without a password.")

    def test_queue_message_on_exchange_with_lxc_container(self):
        """
        If the client is running in an LXC container, the information is
        included in the registration message.
        """
        get_container_info_mock = self.mocker.replace(
            "landscape.lib.vm_info.get_container_info")
        get_container_info_mock()
        self.mocker.result("lxc")
        self.mocker.replay()
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        self.assertEqual("lxc", messages[0]["container-info"])

    def test_queue_message_on_exchange_with_password(self):
        """If a registration password is available, we pass it on!"""
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.config.registration_key = "SEKRET"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        password = messages[0]["registration_password"]
        self.assertEqual("SEKRET", password)
        self.assertEqual(self.logfile.getvalue().strip(),
                         "INFO: Queueing message to register with account "
                         "'account_name' with a password.")

    def test_queue_message_on_exchange_with_tags(self):
        """
        If the admin has defined tags for this computer, we send them to the
        server.
        """
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.config.registration_key = "SEKRET"
        self.config.tags = u"computer,tag"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        self.assertEqual("computer,tag", messages[0]["tags"])
        self.assertEqual(self.logfile.getvalue().strip(),
                         "INFO: Queueing message to register with account "
                         "'account_name' and tags computer,tag "
                         "with a password.")

    def test_queue_message_on_exchange_with_invalid_tags(self):
        """
        If the admin has defined tags for this computer, but they are not
        valid, we drop them, and report an error.
        """
        self.log_helper.ignore_errors("Invalid tags provided for cloud "
                                      "registration")
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.config.registration_key = "SEKRET"
        self.config.tags = u"<script>alert()</script>"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        self.assertIs(None, messages[0]["tags"])
        self.assertEqual(self.logfile.getvalue().strip(),
                         "ERROR: Invalid tags provided for cloud "
                         "registration.\n    "
                         "INFO: Queueing message to register with account "
                         "'account_name' with a password.")

    def test_queue_message_on_exchange_with_unicode_tags(self):
        """
        If the admin has defined tags for this computer, we send them to the
        server.
        """
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.config.registration_key = "SEKRET"
        self.config.tags = u"prova\N{LATIN SMALL LETTER J WITH CIRCUMFLEX}o"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        expected = u"prova\N{LATIN SMALL LETTER J WITH CIRCUMFLEX}o"
        self.assertEqual(expected, messages[0]["tags"])

        self.assertEqual(self.logfile.getvalue().strip(),
                         "INFO: Queueing message to register with account "
                         "'account_name' and tags prova\xc4\xb5o "
                         "with a password.")

    def test_queue_message_on_exchange_with_access_group(self):
        """
        If the admin has defined an access_group for this computer, we send
        it to the server.
        """
        self.mstore.set_accepted_types(["register"])
        self.config.account_name = "account_name"
        self.config.access_group = u"dinosaurs"
        self.config.tags = u"server,london"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        self.assertEqual("dinosaurs", messages[0]["access_group"])
        self.assertEqual(self.logfile.getvalue().strip(),
                         "INFO: Queueing message to register with account "
                         "'account_name' in access group 'dinosaurs' and "
                         "tags server,london without a password.")

    def test_queue_message_on_exchange_with_empty_access_group(self):
        """
        If the access_group is "", then the outgoing message does not define
        an "access_group" key.
        """
        self.mstore.set_accepted_types(["register"])
        self.config.access_group = u""
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        # Make sure the key does not appear in the outgoing message.
        self.assertNotIn("access_group", messages[0])

    def test_queue_message_on_exchange_with_none_access_group(self):
        """
        If the access_group is None, then the outgoing message does not define
        an "access_group" key.
        """
        self.mstore.set_accepted_types(["register"])
        self.config.access_group = None
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        # Make sure the key does not appear in the outgoing message.
        self.assertNotIn("access_group", messages[0])

    def test_queueing_registration_message_resets_message_store(self):
        """
        When a registration message is queued, the store is reset
        entirely, since everything else that was queued is meaningless
        now that we're trying to register again.
        """
        self.mstore.set_accepted_types(["register", "test"])
        self.mstore.add({"type": "test"})
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0]["type"], "register")

    def test_no_message_when_should_register_is_false(self):
        """If we already have a secure id, do not queue a register message.
        """
        handler_mock = self.mocker.patch(self.handler)
        handler_mock.should_register()
        self.mocker.result(False)

        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"

        # If we didn't fake it, it'd work.  We do that to ensure that
        # all the needed data is in place, and that this method is
        # really what decides if a message is sent or not.  This way
        # we can test it individually.
        self.assertTrue(self.handler.should_register())

        # Now let's see.
        self.mocker.replay()

        self.reactor.fire("pre-exchange")
        self.assertMessages(self.mstore.get_pending_messages(), [])

    def test_registration_failed_event(self):
        """
        The deferred returned by a registration request should fail
        with L{InvalidCredentialsError} if the server responds with a
        failure message.
        """
        reactor_mock = self.mocker.patch(self.reactor)
        reactor_mock.fire("registration-failed")
        self.mocker.replay()
        self.exchanger.handle_message(
            {"type": "registration", "info": "unknown-account"})

    def test_registration_failed_event_not_fired_when_uncertain(self):
        """
        If the data in the registration message isn't what we expect,
        the event isn't fired.
        """
        reactor_mock = self.mocker.patch(self.reactor)
        reactor_mock.fire("registration-failed")
        self.mocker.count(0)
        self.mocker.replay()
        self.exchanger.handle_message(
            {"type": "registration", "info": "blah-blah"})

    def test_register_resets_ids(self):
        self.identity.secure_id = "foo"
        self.identity.insecure_id = "bar"
        self.handler.register()
        self.assertEqual(self.identity.secure_id, None)
        self.assertEqual(self.identity.insecure_id, None)

    def test_register_calls_urgent_exchange(self):
        exchanger_mock = self.mocker.patch(self.exchanger)
        exchanger_mock.exchange()
        self.mocker.passthrough()
        self.mocker.replay()
        self.handler.register()

    def test_register_deferred_called_on_done(self):
        # We don't want informational messages.
        self.logger.setLevel(logging.WARNING)

        calls = [0]
        d = self.handler.register()

        def add_call(result):
            self.assertEqual(result, None)
            calls[0] += 1

        d.addCallback(add_call)

        # This should somehow callback the deferred.
        self.exchanger.handle_message(
            {"type": "set-id", "id": "abc", "insecure-id": "def"})

        self.assertEqual(calls, [1])

        # Doing it again to ensure that the deferred isn't called twice.
        self.exchanger.handle_message(
            {"type": "set-id", "id": "abc", "insecure-id": "def"})

        self.assertEqual(calls, [1])

        self.assertEqual(self.logfile.getvalue(), "")

    def test_resynchronize_fired_when_registration_done(self):
        """
        When we call C{register} this should trigger a "resynchronize-clients"
        event with global scope.
        """
        results = []

        def append(scopes=None):
            results.append(scopes)

        self.reactor.call_on("resynchronize-clients", append)

        self.handler.register()

        # This should somehow callback the deferred.
        self.exchanger.handle_message(
            {"type": "set-id", "id": "abc", "insecure-id": "def"})

        self.assertEqual(results, [None])

    def test_register_deferred_called_on_failed(self):
        # We don't want informational messages.
        self.logger.setLevel(logging.WARNING)

        calls = [0]
        d = self.handler.register()

        def add_call(failure):
            exception = failure.value
            self.assertTrue(isinstance(exception, InvalidCredentialsError))
            calls[0] += 1

        d.addErrback(add_call)

        # This should somehow callback the deferred.
        self.exchanger.handle_message(
            {"type": "registration", "info": "unknown-account"})

        self.assertEqual(calls, [1])

        # Doing it again to ensure that the deferred isn't called twice.
        self.exchanger.handle_message(
            {"type": "registration", "info": "unknown-account"})

        self.assertEqual(calls, [1])

        self.assertEqual(self.logfile.getvalue(), "")

    def test_exchange_done_calls_exchange(self):
        exchanger_mock = self.mocker.patch(self.exchanger)
        exchanger_mock.exchange()
        self.mocker.passthrough()
        self.mocker.replay()

        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.reactor.fire("exchange-done")

    def test_exchange_done_wont_call_exchange_when_just_tried(self):
        exchanger_mock = self.mocker.patch(self.exchanger)
        exchanger_mock.exchange()
        self.mocker.count(0)
        self.mocker.replay()

        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.reactor.fire("pre-exchange")
        self.reactor.fire("exchange-done")

    def test_default_hostname(self):
        self.mstore.set_accepted_types(["register"])
        self.config.computer_title = "Computer Title"
        self.config.account_name = "account_name"
        self.config.registration_key = "SEKRET"
        self.reactor.fire("pre-exchange")
        messages = self.mstore.get_pending_messages()
        self.assertEqual(socket.getfqdn(), messages[0]["hostname"])


class JujuRegistrationHandlerTest(RegistrationHandlerTestBase):

    juju_contents = json.dumps({"environment-uuid": "DEAD-BEEF",
                                "unit-name": "service/0",
                                "api-addresses": "10.0.3.1:17070",
                                "private-address": "127.0.0.1"})

    def test_juju_information_added_when_present(self):
        """
        When Juju information is found in $data_dir/juju-info.json,
        key parts of it are sent in the registration message.
        """
        self.mstore.set_accepted_types(["register"])
        self.config.account_name = "account_name"
        self.reactor.fire("run")
        self.reactor.fire("pre-exchange")

        messages = self.mstore.get_pending_messages()
        expected = {"environment-uuid": "DEAD-BEEF",
                    "api-addresses": ["10.0.3.1:17070"],
                    "unit-name": "service/0"}
        self.assertEqual(expected, messages[0]["juju-info"])


class CloudRegistrationHandlerTest(RegistrationHandlerTestBase):

    cloud = True

    def setUp(self):
        super(CloudRegistrationHandlerTest, self).setUp()
        self.query_results = {}

        def fetch_stub(url):
            value = self.query_results[url]
            if isinstance(value, Exception):
                return fail(value)
            else:
                return succeed(value)

        self.fetch_func = fetch_stub

    def get_user_data(self, otps=None,
                      exchange_url="https://example.com/message-system",
                      ping_url="http://example.com/ping",
                      ssl_ca_certificate=None):
        if otps is None:
            otps = ["otp1"]
        user_data = {"otps": otps, "exchange-url": exchange_url,
                     "ping-url": ping_url}
        if ssl_ca_certificate is not None:
            user_data["ssl-ca-certificate"] = ssl_ca_certificate
        return user_data

    def prepare_query_results(
        self, user_data=None, instance_key="key1", launch_index=0,
        local_hostname="ooga.local", public_hostname="ooga.amazon.com",
        reservation_key=u"res1", ramdisk_key=u"ram1", kernel_key=u"kernel1",
        image_key=u"image1", local_ip="10.0.0.1", public_ip="10.0.0.2",
        ssl_ca_certificate=None):
        if user_data is None:
            user_data = self.get_user_data(
                ssl_ca_certificate=ssl_ca_certificate)
        if not isinstance(user_data, Exception):
            user_data = dumps(user_data)
        api_base = "http://169.254.169.254/latest"
        self.query_results.clear()
        for url_suffix, value in [
            ("/user-data", user_data),
            ("/meta-data/instance-id", instance_key),
            ("/meta-data/reservation-id", reservation_key),
            ("/meta-data/local-hostname", local_hostname),
            ("/meta-data/public-hostname", public_hostname),
            ("/meta-data/ami-launch-index", str(launch_index)),
            ("/meta-data/kernel-id", kernel_key),
            ("/meta-data/ramdisk-id", ramdisk_key),
            ("/meta-data/ami-id", image_key),
            ("/meta-data/local-ipv4", local_ip),
            ("/meta-data/public-ipv4", public_ip),
            ]:
            self.query_results[api_base + url_suffix] = value

    def prepare_cloud_registration(self, account_name=None,
                                   registration_key=None, tags=None,
                                   access_group=None):
        # Set things up so that the client thinks it should register
        self.mstore.set_accepted_types(list(self.mstore.get_accepted_types())
                                       + ["register-cloud-vm"])
        self.config.account_name = account_name
        self.config.registration_key = registration_key
        self.config.computer_title = None
        self.config.tags = tags
        self.config.access_group = access_group
        self.identity.secure_id = None
        self.assertTrue(self.handler.should_register())

    def get_expected_cloud_message(self, **kwargs):
        """
        Return the message which is expected from a similar call to
        L{get_registration_handler_for_cloud}.
        """
        message = dict(type="register-cloud-vm",
                       otp="otp1",
                       hostname="ooga.local",
                       local_hostname="ooga.local",
                       public_hostname="ooga.amazon.com",
                       instance_key=u"key1",
                       reservation_key=u"res1",
                       ramdisk_key=u"ram1",
                       kernel_key=u"kernel1",
                       launch_index=0,
                       image_key=u"image1",
                       account_name=None,
                       registration_password=None,
                       local_ipv4=u"10.0.0.1",
                       public_ipv4=u"10.0.0.2",
                       tags=None)
        # The get_vm_info() needs to be deferred to the else.  If vm-info is
        # not specified in kwargs, get_vm_info() will typically be mocked.
        if "vm_info" in kwargs:
            message["vm-info"] = kwargs.pop("vm_info")
        else:
            message["vm-info"] = get_vm_info()
        message.update(kwargs)
        return message

    def test_cloud_registration(self):
        """
        When the 'cloud' configuration variable is set, cloud registration is
        done instead of normal password-based registration. This means:

        - "Launch Data" is fetched from the EC2 Launch Data URL. This contains
          a one-time password that is used during registration.
        - A different "register-cloud-vm" message is sent to the server instead
          of "register", containing the OTP. This message is handled by
          immediately accepting the computer, instead of going through the
          pending computer stage.
        """
        get_vm_info_mock = self.mocker.replace(get_vm_info)
        get_vm_info_mock()
        self.mocker.result("xen")
        self.mocker.replay()
        self.prepare_query_results()
        self.prepare_cloud_registration(tags=u"server,london")

        # metadata is fetched and stored at reactor startup:
        self.reactor.fire("run")

        # And the metadata returned determines the URLs that are used
        self.assertEqual(self.transport.get_url(),
                         "https://example.com/message-system")
        self.assertEqual(self.pinger.get_url(),
                         "http://example.com/ping")
        # Lets make sure those values were written back to the config file
        new_config = BrokerConfiguration()
        new_config.load_configuration_file(self.config_filename)
        self.assertEqual(new_config.url, "https://example.com/message-system")
        self.assertEqual(new_config.ping_url, "http://example.com/ping")

        # Okay! Exchange should cause the registration to happen.
        self.exchanger.exchange()
        # This *should* be asynchronous, but I think a billion tests are
        # written like this
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(
            self.transport.payloads[0]["messages"],
            [self.get_expected_cloud_message(tags=u"server,london",
                                             vm_info="xen")])

    def test_cloud_registration_with_access_group(self):
        """
        If the access_group field is presnet in the configuration, the
        access_group field is present in the outgoing message for a VM
        registration, and a notice appears in the logs.
        """
        self.prepare_query_results()
        self.prepare_cloud_registration(access_group=u"dinosaurs",
                                        tags=u"server,london")

        self.reactor.fire("run")
        self.exchanger.exchange()
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(
            self.transport.payloads[0]["messages"],
            [self.get_expected_cloud_message(
                access_group=u"dinosaurs", tags=u"server,london")])

    def test_cloud_registration_with_otp(self):
        """
        If the OTP is present in the configuration, it's used to trigger the
        registration instead of using the user data.
        """
        self.config.otp = "otp1"
        self.prepare_query_results(user_data=None)

        self.prepare_cloud_registration()

        # metadata is fetched and stored at reactor startup:
        self.reactor.fire("run")

        # Okay! Exchange should cause the registration to happen.
        self.exchanger.exchange()
        # This *should* be asynchronous, but I think a billion tests are
        # written like this
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(
            self.transport.payloads[0]["messages"],
            [self.get_expected_cloud_message()])

    def test_cloud_registration_with_invalid_tags(self):
        """
        Invalid tags in the configuration should result in the tags not being
        sent to the server, and this fact logged.
        """
        self.log_helper.ignore_errors("Invalid tags provided for cloud "
                                      "registration")
        self.prepare_query_results()
        self.prepare_cloud_registration(tags=u"<script>alert()</script>,hardy")

        # metadata is fetched and stored at reactor startup:
        self.reactor.fire("run")
        self.exchanger.exchange()
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(self.transport.payloads[0]["messages"],
                            [self.get_expected_cloud_message(tags=None)])
        self.assertEqual(self.logfile.getvalue().strip()[:-7],
                         "ERROR: Invalid tags provided for cloud "
                         "registration.\n    "
                         "INFO: Queueing message to register with OTP\n    "
                         "INFO: Starting message exchange with "
                         "https://example.com/message-system.\n    "
                         "INFO: Message exchange completed in")

    def test_cloud_registration_with_ssl_ca_certificate(self):
        """
        If we have an SSL certificate CA included in the user-data, this should
        be written out, and the configuration updated to reflect this.
        """
        key_filename = os.path.join(self.config.data_path,
            "%s.ssl_public_key" % os.path.basename(self.config_filename))

        print_text_mock = self.mocker.replace(print_text)
        print_text_mock("Writing SSL CA certificate to %s..." %
                        key_filename)
        self.mocker.replay()
        self.prepare_query_results(ssl_ca_certificate=u"1234567890")
        self.prepare_cloud_registration(tags=u"server,london")
        # metadata is fetched and stored at reactor startup:
        self.reactor.fire("run")
        # And the metadata returned determines the URLs that are used
        self.assertEqual("https://example.com/message-system",
                         self.transport.get_url())
        self.assertEqual(key_filename, self.transport._pubkey)
        self.assertEqual("http://example.com/ping",
                         self.pinger.get_url())
        # Let's make sure those values were written back to the config file
        new_config = BrokerConfiguration()
        new_config.load_configuration_file(self.config_filename)
        self.assertEqual("https://example.com/message-system", new_config.url)
        self.assertEqual("http://example.com/ping", new_config.ping_url)
        self.assertEqual(key_filename, new_config.ssl_public_key)
        self.assertEqual("1234567890", open(key_filename, "r").read())

    def test_wrong_user_data(self):
        self.prepare_query_results(user_data="other stuff, not a bpickle")
        self.prepare_cloud_registration()

        # Mock registration-failed call
        reactor_mock = self.mocker.patch(self.reactor)
        reactor_mock.fire("registration-failed")
        self.mocker.replay()

        self.reactor.fire("run")
        self.exchanger.exchange()

    def test_wrong_object_type_in_user_data(self):
        self.prepare_query_results(user_data=True)
        self.prepare_cloud_registration()

        # Mock registration-failed call
        reactor_mock = self.mocker.patch(self.reactor)
        reactor_mock.fire("registration-failed")
        self.mocker.replay()

        self.reactor.fire("run")
        self.exchanger.exchange()

    def test_user_data_with_not_enough_elements(self):
        """
        If the AMI launch index isn't represented in the list of OTPs in the
        user data then BOOM.
        """
        self.prepare_query_results(launch_index=1)
        self.prepare_cloud_registration()

        # Mock registration-failed call
        reactor_mock = self.mocker.patch(self.reactor)
        reactor_mock.fire("registration-failed")
        self.mocker.replay()

        self.reactor.fire("run")
        self.exchanger.exchange()

    def test_user_data_bpickle_without_otp(self):
        self.prepare_query_results(user_data={"foo": "bar"})
        self.prepare_cloud_registration()

        # Mock registration-failed call
        reactor_mock = self.mocker.patch(self.reactor)
        reactor_mock.fire("registration-failed")
        self.mocker.replay()

        self.reactor.fire("run")
        self.exchanger.exchange()

    def test_no_otp_fallback_to_account(self):
        self.prepare_query_results(user_data="other stuff, not a bpickle",
                                   instance_key=u"key1")
        self.prepare_cloud_registration(account_name=u"onward",
                                        registration_key=u"password",
                                        tags=u"london,server")

        self.reactor.fire("run")
        self.exchanger.exchange()

        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(self.transport.payloads[0]["messages"],
                            [self.get_expected_cloud_message(
                                otp=None,
                                account_name=u"onward",
                                registration_password=u"password",
                                tags=u"london,server")])
        self.assertEqual(self.logfile.getvalue().strip()[:-7],
           "INFO: Queueing message to register with account u'onward' and "
           "tags london,server as an EC2 instance.\n    "
           "INFO: Starting message exchange with http://localhost:91919.\n    "
           "INFO: Message exchange completed in")

    def test_queueing_cloud_registration_message_resets_message_store(self):
        """
        When a registration from a cloud is about to happen, the message store
        is reset, because all previous messages are now meaningless.
        """
        self.mstore.set_accepted_types(list(self.mstore.get_accepted_types())
                                       + ["test"])

        self.mstore.add({"type": "test"})

        self.prepare_query_results()

        self.prepare_cloud_registration()

        self.reactor.fire("run")
        self.reactor.fire("pre-exchange")

        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0]["type"], "register-cloud-vm")

    def test_cloud_registration_fetch_errors(self):
        """
        If fetching metadata fails, and we have no account details to fall
        back to, we fire 'registration-failed'.
        """
        self.log_helper.ignore_errors(pycurl.error)

        def fetch_stub(url):
            return fail(pycurl.error(7, "couldn't connect to host"))

        self.handler = RegistrationHandler(
            self.config, self.identity, self.reactor, self.exchanger,
            self.pinger, self.mstore, fetch_async=fetch_stub)

        self.fetch_stub = fetch_stub
        self.prepare_query_results()
        self.fetch_stub = fetch_stub

        self.prepare_cloud_registration()

        failed = []
        self.reactor.call_on(
            "registration-failed", lambda: failed.append(True))

        self.log_helper.ignore_errors("Got error while fetching meta-data")
        self.reactor.fire("run")
        self.exchanger.exchange()
        self.assertEqual(failed, [True])
        self.assertIn('error: (7, "couldn\'t connect to host")',
                      self.logfile.getvalue())

    def test_cloud_registration_continues_without_user_data(self):
        """
        If no user-data exists (i.e., the user-data URL returns a 404), then
        register-cloud-vm still occurs.
        """
        self.log_helper.ignore_errors(HTTPCodeError)
        self.prepare_query_results(user_data=HTTPCodeError(404, "ohno"))
        self.prepare_cloud_registration(account_name="onward",
                                        registration_key="password")

        self.reactor.fire("run")
        self.exchanger.exchange()
        self.assertIn("HTTPCodeError: Server returned HTTP code 404",
                      self.logfile.getvalue())
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(self.transport.payloads[0]["messages"],
                            [self.get_expected_cloud_message(
                                otp=None,
                                account_name=u"onward",
                                registration_password=u"password")])

    def test_cloud_registration_continues_without_ramdisk(self):
        """
        If the instance doesn't have a ramdisk (ie, the query for ramdisk
        returns a 404), then register-cloud-vm still occurs.
        """
        self.log_helper.ignore_errors(HTTPCodeError)
        self.prepare_query_results(ramdisk_key=HTTPCodeError(404, "ohno"))
        self.prepare_cloud_registration()

        self.reactor.fire("run")
        self.exchanger.exchange()
        self.assertIn("HTTPCodeError: Server returned HTTP code 404",
                      self.logfile.getvalue())
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(self.transport.payloads[0]["messages"],
                            [self.get_expected_cloud_message(
                                ramdisk_key=None)])

    def test_cloud_registration_continues_without_kernel(self):
        """
        If the instance doesn't have a kernel (ie, the query for kernel
        returns a 404), then register-cloud-vm still occurs.
        """
        self.log_helper.ignore_errors(HTTPCodeError)
        self.prepare_query_results(kernel_key=HTTPCodeError(404, "ohno"))
        self.prepare_cloud_registration()

        self.reactor.fire("run")
        self.exchanger.exchange()
        self.assertIn("HTTPCodeError: Server returned HTTP code 404",
                      self.logfile.getvalue())
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(self.transport.payloads[0]["messages"],
                            [self.get_expected_cloud_message(
                                kernel_key=None)])

    def test_fall_back_to_normal_registration_when_metadata_fetch_fails(self):
        """
        If fetching metadata fails, but we do have an account name, then we
        fall back to normal 'register' registration.
        """
        self.mstore.set_accepted_types(["register"])
        self.log_helper.ignore_errors(HTTPCodeError)
        self.prepare_query_results(
            public_hostname=HTTPCodeError(404, "ohnoes"))
        self.prepare_cloud_registration(account_name="onward",
                                        registration_key="password")
        self.config.computer_title = "whatever"
        self.reactor.fire("run")
        self.exchanger.exchange()
        self.assertIn("HTTPCodeError: Server returned HTTP code 404",
                      self.logfile.getvalue())
        self.assertEqual(len(self.transport.payloads), 1)
        messages = self.transport.payloads[0]["messages"]
        self.assertEqual("register", messages[0]["type"])

    def test_should_register_in_cloud(self):
        """
        The client should register when it's in the cloud even though
        it doesn't have the normal account details.
        """
        self.mstore.set_accepted_types(self.mstore.get_accepted_types()
                                       + ("register-cloud-vm",))
        self.config.account_name = None
        self.config.registration_key = None
        self.config.computer_title = None
        self.identity.secure_id = None
        self.assertTrue(self.handler.should_register())

    def test_launch_index(self):
        """
        The client used the value in C{ami-launch-index} to choose the
        appropriate OTP in the user data.
        """
        otp = "correct otp for launch index"
        self.prepare_query_results(
            user_data=self.get_user_data(otps=["wrong index", otp,
                                               "wrong again"],),
            instance_key="key1",
            launch_index=1)

        self.prepare_cloud_registration()

        self.reactor.fire("run")
        self.exchanger.exchange()
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(self.transport.payloads[0]["messages"],
                            [self.get_expected_cloud_message(otp=otp,
                                                             launch_index=1)])

    def test_should_not_register_in_cloud(self):
        """
        Having a secure ID means we shouldn't register, even in the cloud.
        """
        self.mstore.set_accepted_types(self.mstore.get_accepted_types()
                                       + ("register-cloud-vm",))
        self.config.account_name = None
        self.config.registration_key = None
        self.config.computer_title = None
        self.identity.secure_id = "hello"
        self.assertFalse(self.handler.should_register())

    def test_should_not_register_without_register_cloud_vm(self):
        """
        If the server isn't accepting a 'register-cloud-vm' message,
        we shouldn't register.
        """
        self.config.account_name = None
        self.config.registration_key = None
        self.config.computer_title = None
        self.identity.secure_id = None
        self.assertFalse(self.handler.should_register())


class IsCloudManagedTests(LandscapeTest):

    def setUp(self):
        super(IsCloudManagedTests, self).setUp()
        self.urls = []
        self.responses = []

    def fake_fetch(self, url, connect_timeout=None):
        self.urls.append((url, connect_timeout))
        return self.responses.pop(0)

    def mock_socket(self):
        """
        Mock out socket usage by is_cloud_managed to wait for the network.
        """
        # Mock the socket.connect call that it also does
        socket_class = self.mocker.replace("socket.socket", passthrough=False)
        socket = socket_class()
        socket.connect((EC2_HOST, 80))
        socket.close()

    def test_is_managed(self):
        """
        L{is_cloud_managed} returns True if the EC2 user-data contains
        Landscape instance information.  It fetches the EC2 data with low
        timeouts.
        """
        user_data = {"otps": ["otp1"], "exchange-url": "http://exchange",
                     "ping-url": "http://ping"}
        self.responses = [dumps(user_data), "0"]

        self.mock_socket()
        self.mocker.replay()

        self.assertTrue(is_cloud_managed(self.fake_fetch))
        self.assertEqual(
            self.urls,
            [(EC2_API + "/user-data", 5),
             (EC2_API + "/meta-data/ami-launch-index", 5)])

    def test_is_managed_index(self):
        user_data = {"otps": ["otp1", "otp2"],
                     "exchange-url": "http://exchange",
                     "ping-url": "http://ping"}
        self.responses = [dumps(user_data), "1"]
        self.mock_socket()
        self.mocker.replay()
        self.assertTrue(is_cloud_managed(self.fake_fetch))

    def test_is_managed_wrong_index(self):
        user_data = {"otps": ["otp1"], "exchange-url": "http://exchange",
                     "ping-url": "http://ping"}
        self.responses = [dumps(user_data), "1"]
        self.mock_socket()
        self.mocker.replay()
        self.assertFalse(is_cloud_managed(self.fake_fetch))

    def test_is_managed_exchange_url(self):
        user_data = {"otps": ["otp1"], "ping-url": "http://ping"}
        self.responses = [dumps(user_data), "0"]
        self.mock_socket()
        self.mocker.replay()
        self.assertFalse(is_cloud_managed(self.fake_fetch))

    def test_is_managed_ping_url(self):
        user_data = {"otps": ["otp1"], "exchange-url": "http://exchange"}
        self.responses = [dumps(user_data), "0"]
        self.mock_socket()
        self.mocker.replay()
        self.assertFalse(is_cloud_managed(self.fake_fetch))

    def test_is_managed_bpickle(self):
        self.responses = ["some other user data", "0"]
        self.mock_socket()
        self.mocker.replay()
        self.assertFalse(is_cloud_managed(self.fake_fetch))

    def test_is_managed_no_data(self):
        self.responses = ["", "0"]
        self.mock_socket()
        self.mocker.replay()
        self.assertFalse(is_cloud_managed(self.fake_fetch))

    def test_is_managed_fetch_not_found(self):

        def fake_fetch(url, connect_timeout=None):
            raise HTTPCodeError(404, "ohnoes")

        self.mock_socket()
        self.mocker.replay()
        self.assertFalse(is_cloud_managed(fake_fetch))

    def test_is_managed_fetch_error(self):

        def fake_fetch(url, connect_timeout=None):
            raise FetchError(7, "couldn't connect to host")

        self.mock_socket()
        self.mocker.replay()
        self.assertFalse(is_cloud_managed(fake_fetch))

    def test_waits_for_network(self):
        """
        is_cloud_managed will wait until the network before trying to fetch
        the EC2 user data.
        """
        user_data = {"otps": ["otp1"], "exchange-url": "http://exchange",
                     "ping-url": "http://ping"}
        self.responses = [dumps(user_data), "0"]

        self.mocker.order()
        time_sleep = self.mocker.replace("time.sleep", passthrough=False)
        socket_class = self.mocker.replace("socket.socket", passthrough=False)
        socket_obj = socket_class()
        socket_obj.connect((EC2_HOST, 80))
        self.mocker.throw(socket.error("woops"))
        time_sleep(1)
        socket_obj = socket_class()
        socket_obj.connect((EC2_HOST, 80))
        self.mocker.result(None)
        socket_obj.close()
        self.mocker.replay()
        self.assertTrue(is_cloud_managed(self.fake_fetch))

    def test_waiting_times_out(self):
        """
        We'll only wait five minutes for the network to come up.
        """

        def fake_fetch(url, connect_timeout=None):
            raise FetchError(7, "couldn't connect to host")

        self.mocker.order()
        time_sleep = self.mocker.replace("time.sleep", passthrough=False)
        time_time = self.mocker.replace("time.time", passthrough=False)
        time_time()
        self.mocker.result(100)
        socket_class = self.mocker.replace("socket.socket", passthrough=False)
        socket_obj = socket_class()
        socket_obj.connect((EC2_HOST, 80))
        self.mocker.throw(socket.error("woops"))
        time_sleep(1)
        time_time()
        self.mocker.result(401)
        self.mocker.replay()
        # Mocking time.time is dangerous, because the test harness calls it. So
        # we explicitly reset mocker before returning from the test.
        try:
            self.assertFalse(is_cloud_managed(fake_fetch))
        finally:
            self.mocker.reset()


class ProvisioningRegistrationTest(RegistrationHandlerTestBase):

    def test_provisioned_machine_registration_with_otp(self):
        """
        Register provisioned machines using an OTP.
        """
        self.mstore.set_accepted_types(["register-provisioned-machine"])
        self.config.account_name = ""
        self.config.provisioning_otp = "ohteepee"
        self.reactor.fire("pre-exchange")

        self.assertMessages([{"otp": "ohteepee", "timestamp": 0, "api": "3.2",
                              "type": "register-provisioned-machine"}],
                            self.mstore.get_pending_messages())
        self.assertEqual(u"INFO: Queueing message to register with OTP as a"
                         u" newly provisioned machine.",
                         self.logfile.getvalue().strip())

        self.exchanger.exchange()
        self.assertMessages([{"otp": "ohteepee", "timestamp": 0, "api": "3.2",
                              "type": "register-provisioned-machine"}],
                            self.transport.payloads[0]["messages"])

    def test_provisioned_machine_registration_with_empty_otp(self):
        """
        No message should be sent when an empty OTP is passed.
        """
        self.mstore.set_accepted_types(["register-provisioned-machine"])
        self.config.account_name = ""
        self.config.provisioning_otp = ""
        self.reactor.fire("pre-exchange")

        self.assertMessages([], self.mstore.get_pending_messages())
        self.assertEqual(u"", self.logfile.getvalue().strip())



from __future__ import division

import os

from twisted.internet.defer import succeed

from landscape.lib.disk import (get_mount_info, get_filesystem_for_path)


def format_megabytes(megabytes):
    if megabytes >= 1024 * 1024:
        return "%.2fTB" % (megabytes / (1024 * 1024))
    elif megabytes >= 1024:
        return "%.2fGB" % (megabytes / 1024)
    else:
        return "%dMB" % (megabytes)


def usage(info):
    total = info["total-space"]
    used = total - info["free-space"]
    return "%0.1f%% of %s" % ((used / total) * 100, format_megabytes(total))


class Disk(object):

    def __init__(self, mounts_file="/proc/mounts", statvfs=os.statvfs):
        self._mounts_file = mounts_file
        self._statvfs = statvfs

    def register(self, sysinfo):
        self._sysinfo = sysinfo

    def run(self):
        main_info = get_filesystem_for_path("/home", self._mounts_file,
                                            self._statvfs)
        if main_info is not None:
            total = main_info["total-space"]
            if total <= 0:
                root_main_info = get_filesystem_for_path(
                    "/", self._mounts_file, self._statvfs)
                if root_main_info is not None:
                    total = root_main_info["total-space"]
                    main_info = root_main_info
            if total <= 0:
                main_usage = "unknown"
            else:
                main_usage = usage(main_info)
            self._sysinfo.add_header("Usage of " + main_info["mount-point"],
                                     main_usage)
        else:
            self._sysinfo.add_header("Usage of /home", "unknown")

        seen_mounts = set()
        seen_devices = set()
        infos = list(get_mount_info(self._mounts_file, self._statvfs))
        infos.sort(key=lambda i: len(i["mount-point"]))
        for info in infos:
            total = info["total-space"]
            mount_seen = info["mount-point"] in seen_mounts
            device_seen = info["device"] in seen_devices
            seen_mounts.add(info["mount-point"])
            seen_devices.add(info["device"])
            if mount_seen or device_seen:
                continue

            if total <= 0:
                # Some "virtual" filesystems have 0 total space. ignore them.
                continue

            used = ((total - info["free-space"]) / total) * 100
            if used >= 85:
                self._sysinfo.add_note("%s is using %s"
                                       % (info["mount-point"], usage(info)))
        return succeed(None)



from landscape.diff import diff


class UserChanges(object):
    """Detect changes made since the last snapshot was taken.

    If no snapshot is available all users and groups are reported.
    When a snapshot is available, only the changes between the current
    state and the snapshotted state are transmitted to the server.
    """

    def __init__(self, persist, provider):
        super(UserChanges, self).__init__()
        self._persist = persist
        self._provider = provider
        # FIXME This shouldn't really be necessary.  Not having it
        # here with the current factoring is also problematic.  Figure
        # out a clean way to factor this.  Gustavo suggested splitting
        # it into _build_old_data and _build_new_data and just calling
        # that from the necessary places.
        self._refresh()

    def _refresh(self):
        """Load the previous snapshot and update current data."""
        self._old_users = self._persist.get("users", {})
        self._old_groups = self._persist.get("groups", {})
        self._new_users = self._create_index(
            "username", self._provider.get_users())
        self._new_groups = self._create_index(
            "name", self._provider.get_groups())

    def snapshot(self):
        """Save the current state and use it as a comparison snapshot."""
        self._persist.set("users", self._new_users)
        self._persist.set("groups", self._new_groups)

    def clear(self):
        """
        Reset the snapshot state and forget all knowledge of users and
        groups.
        """
        self._persist.remove("users")
        self._persist.remove("groups")

    def _create_index(self, key, sequence):
        """
        Given a key and a sequence of dicts, return a dict of the form
        C{{dict[key]: dict, ...}}.
        """
        index = {}
        for data in sequence:
            index[data[key]] = data
        return index

    def create_diff(self):
        """Returns the changes since the last snapshot.
        
        See landscape.message_schemas.USERS schema for a description of the
        dictionary returned by this method.
        """
        self._refresh()
        changes = {}
        changes.update(self._detect_user_changes())
        changes.update(self._detect_group_changes())
        return changes

    def _detect_user_changes(self):
        """
        Compare the current user snapshot to the old one and return a
        C{dict} with C{create-users}, C{update-users} and
        C{delete-users} fields.  Fields without data aren't included
        in the result.
        """
        changes = {}
        creates, updates, deletes = diff(self._old_users, self._new_users)
        if creates:
            changes["create-users"] = list(creates.itervalues())
        if updates:
            changes["update-users"] = list(updates.itervalues())
        if deletes:
            changes["delete-users"] = list(deletes.iterkeys())
        return changes

    def _detect_group_changes(self):
        """
        Compare the current group snapshot to the old one and create a
        C{dict} with C{create-groups}, C{delete-groups},
        C{create-group-members} and {delete-group-members} fields.
        Fields without data aren't included in the result.
        """
        changes = {}
        creates, updates, deletes = diff(self._old_groups, self._new_groups)

        if creates:
            groups = []
            create_members = {}
            for value in creates.itervalues():
                # Use a copy to avoid removing the 'members' element
                # from stored data.
                value = value.copy()
                members = value.pop("members")
                if members:
                    create_members[value["name"]] = members
                groups.append(value)
            changes["create-groups"] = groups
            if create_members:
                changes["create-group-members"] = create_members

        if updates:
            remove_members = {}
            create_members = {}
            update_groups = []
            for groupname, new_data in updates.iteritems():
                old_data = self._old_groups[groupname]
                old_members = set(old_data["members"])
                new_members = set(new_data["members"])
                created = new_members - old_members
                if created:
                    create_members[groupname] = sorted(created)
                removed = old_members - new_members
                if removed:
                    remove_members[groupname] = sorted(removed)
                if old_data["gid"] != new_data["gid"]:
                    update_groups.append({"name": groupname,
                                          "gid": new_data["gid"]})
            if create_members:
                members = changes.setdefault("create-group-members", {})
                members.update(create_members)
            if remove_members:
                members = changes.setdefault("delete-group-members", {})
                members.update(remove_members)
            if update_groups:
                members = changes.setdefault("update-groups", [])
                members.extend(update_groups)

        if deletes:
            changes["delete-groups"] = deletes.keys()

        return changes



import logging
import time
import os
import json

from landscape.lib.fetch import fetch, HTTPCodeError, PyCurlError, FetchError
from landscape.lib.monitor import CoverageMonitor
from landscape.lib.network import get_active_device_info
from landscape.monitor.plugin import MonitorPlugin


class SwiftDeviceInfo(MonitorPlugin):

    persist_name = "swift-device-info"
    scope = "storage"

    def __init__(self, interval=300, monitor_interval=60 * 60,
                 create_time=time.time,
                 swift_config="/etc/swift/object-server.conf",
                 swift_ring="/etc/swift/object.ring.gz"):
        self.run_interval = interval
        self._monitor_interval = monitor_interval
        self._create_time = create_time
        self._fetch = fetch
        self._get_network_devices = get_active_device_info
        self._swift_config = swift_config  # If exists, we are a swift node
        self._swift_ring = swift_ring      # To discover swift recon port
        self._swift_recon_url = None
        self._create_time = create_time
        self._swift_device_info = []
        self._swift_device_info_to_persist = []
        self.enabled = True

    def register(self, registry):
        super(SwiftDeviceInfo, self).register(registry)
        self._monitor = CoverageMonitor(self.run_interval, 0.8,
                                        "swift device info snapshot",
                                        create_time=self._create_time)
        self.registry.reactor.call_every(self._monitor_interval,
                                         self._monitor.log)
        self.registry.reactor.call_on("stop", self._monitor.log, priority=2000)
        self.call_on_accepted("swift-device-info", self.send_messages, True)

    def create_swift_device_info_message(self):
        if self._swift_device_info:
            message = {"type": "swift-device-info",
                       "swift-device-info": self._swift_device_info}
            self._swift_device_info_to_persist = self._swift_device_info[:]
            self._swift_device_info = []
            return message
        return None

    def send_messages(self, urgent=False):
        message = self.create_swift_device_info_message()
        if message:
            logging.info("Queueing message with updated swift device info.")
            d = self.registry.broker.send_message(
                message, self._session_id, urgent=urgent)
            d.addCallback(lambda x: self.persist_swift_info())

    def exchange(self):
        self.registry.broker.call_if_accepted("swift-device-info",
                                              self.send_messages)

    def persist_swift_info(self):
        for swift_device_info in self._swift_device_info_to_persist:
            device_name = swift_device_info["device"]
            key = (self.persist_name, device_name)
            self._persist.set(key, swift_device_info)
        self._swift_device_info_to_persist = None
        # This forces the registry to write the persistent store to disk
        # This means that the persistent data reflects the state of the
        # messages sent.
        self.registry.flush()

    def run(self):
        if not self.enabled:
            return
        self._monitor.ping()

        current_swift_devices = self._get_swift_devices()
        current_device_names = []
        for swift_info in current_swift_devices:
            device_name = swift_info["device"]
            current_device_names.append(device_name)
            key = (self.persist_name, device_name)
            prev_swift_info = self._persist.get(key)
            if not prev_swift_info or prev_swift_info != swift_info:
                if swift_info not in self._swift_device_info:
                    self._swift_device_info.append(swift_info)

        # Get all persisted devices and remove those that no longer exist
        persisted_devices = self._persist.get(self.persist_name)
        if persisted_devices:
            for device_name in persisted_devices.keys():
                if device_name not in current_device_names:
                    self._persist.remove((self.persist_name, device_name))

    def _get_swift_devices(self):
        config_file = self._swift_config
        # Check if a swift storage config file is available. No need to run
        # if we know that we're not on a swift monitor node anyway.
        if not os.path.exists(config_file):
            # There is no config file - it's not a swift storage machine.
            self.enabled = False
            logging.info(
                    "This does not appear to be a swift storage server. '%s' "
                    "plugin has been disabled." % self.persist_name)
            return []

        # Extract the swift service URL from the ringfile and cache it.
        if self._swift_recon_url is None:
            ring = self._get_ring()
            if ring is None:
                return []

            network_devices = self._get_network_devices()
            local_ips = [device["ip_address"] for device in network_devices]

            # Grab first swift service with an IP on this host
            for dev in ring.devs:
                if dev and dev["ip"] in local_ips:
                    self._swift_recon_url = "http://%s:%d/recon/diskusage" % (
                        dev['ip'], dev['port'])
                    break

            if self._swift_recon_url is None:
                self.enabled = False
                logging.error(
                    "Local swift service not found. '%s' plugin has "
                    "been disabled." % self.persist_name)
                return []

        recon_disk_info = self._get_swift_disk_usage()
        # We don't care about avail and free figures because we track
        # free_space for mounted devices in free-space messages
        return [{"device": "/dev/%s" % device["device"],
                 "mounted":  device["mounted"]} for device in recon_disk_info]

    def _get_swift_disk_usage(self):
        """
        Query the swift storage usage data by parsing the curled recon data
        from http://localhost:<_swift_service_port>/recon/diskusage.
        Lots of recon data for the picking described at:
        http://docs.openstack.org/developer/swift/admin_guide.html
        """
        error_message = None
        try:
            content = self._fetch(self._swift_recon_url)
        except HTTPCodeError, error:
            error_message = (
                "Swift service is running without swift-recon enabled.")
        except (FetchError, PyCurlError), error:
            error_message = (
                "Swift service not available at %s. %s." %
                (self._swift_recon_url, str(error)))
        if error_message is not None:
            self.enabled = False
            logging.error("%s '%s' plugin has been disabled." % (
                error_message, self.persist_name))
            return None

        if not content:
            return None

        swift_disk_usages = json.loads(content)  # list of device dicts
        return swift_disk_usages

    def _get_ring(self):
        """Return ring-file object from self._swift_ring location"""
        if not os.path.exists(self._swift_ring):
            logging.warning(
                "Swift ring files are not available yet.")
            return None
        try:
            from swift.common.ring import Ring
        except ImportError:
            self.enabled = False
            logging.error(
                "Swift python common libraries not found. '%s' plugin has "
                "been disabled." % self.persist_name)
            return None
        return Ring(self._swift_ring)



import re
import os

from gettext import gettext as _

from gi.repository import GObject, Gtk

from landscape.ui.constants import (
    CANONICAL_MANAGED, LOCAL_MANAGED, NOT_MANAGED)

# Note, I think this may not be fully compliant with the changes in RFC 1123
HOST_NAME_REGEXP = re.compile("^(([a-zA-Z][a-zA-Z0-9\-]*)?[a-zA-Z0-9][\.]?)*"
                              "(([A-Za-z][A-Za-z0-9\-]*)?[A-Za-z0-9])$")


def sanitise_host_name(host_name):
    """
    Do some minimal host name sanitation.
    """
    return host_name.strip()


def is_valid_host_name(host_name):
    """
    Check that the provided host name complies with L{HOST_NAME_REGEXP} and is
    therefor valid.
    """
    return HOST_NAME_REGEXP.match(host_name) is not None


def is_ascii(text):
    """
    Test that the provided string contains only characters from the ASCII set.
    """
    try:
        text.decode("ascii")
        return True
    except UnicodeDecodeError:
        return False


class ClientSettingsDialog(Gtk.Dialog):
    """
    L{ClientSettingsDialog} is a subclass of Gtk.Dialog that loads the UI
    components from the associated Glade XML file and wires everything up to
    the controller.
    """

    GLADE_FILE = "landscape-client-settings.glade"
    INVALID_HOST_NAME = 0
    UNICODE_IN_ENTRY = 1

    def __init__(self, controller):
        super(ClientSettingsDialog, self).__init__(
            title=_("Landscape Service"),
            flags=Gtk.DialogFlags.MODAL)
        self.set_default_icon_name("preferences-management-service")
        self.set_resizable(False)
        self._initialised = False
        self._validation_errors = set()
        self._errored_entries = []
        self.controller = controller
        self.setup_ui()
        self.load_data()
        # One extra revert to reset after loading data
        self.controller.revert()

    def indicate_error_on_entry(self, entry):
        """
        Show a warning icon on a L{Gtk.Entry} to indicate some associated
        error.
        """
        entry.set_icon_from_stock(
            Gtk.EntryIconPosition.PRIMARY, Gtk.STOCK_DIALOG_WARNING)
        self._errored_entries.append(entry)

    def check_local_landscape_host_name_entry(self):
        host_name = sanitise_host_name(
            self.local_landscape_host_entry.get_text())
        ascii_ok = is_ascii(host_name)
        host_name_ok = is_valid_host_name(host_name)
        if ascii_ok and host_name_ok:
            self.local_landscape_host_entry.set_text(host_name)
            return True
        else:
            self.indicate_error_on_entry(self.local_landscape_host_entry)
            if not host_name_ok:
                self._validation_errors.add(self.INVALID_HOST_NAME)
            if not ascii_ok:
                self._validation_errors.add(self.UNICODE_IN_ENTRY)
            return False

    def check_entry(self, entry):
        """
        Check that the text content of a L{Gtk.Entry} is acceptable.
        """
        if is_ascii(entry.get_text()):
            return True
        else:
            self.indicate_error_on_entry(entry)
            self._validation_errors.add(self.UNICODE_IN_ENTRY)
            return False

    def validity_check(self):
        self._validation_errors = set()
        if self._info_bar_container.get_visible():
            self.dismiss_infobar(None)
        active_iter = self.liststore.get_iter(
            self.use_type_combobox.get_active())
        [management_type] = self.liststore.get(active_iter, 0)
        if management_type == NOT_MANAGED:
            return True
        elif management_type == CANONICAL_MANAGED:
            account_name_ok = self.check_entry(self.hosted_account_name_entry)
            password_ok = self.check_entry(self.hosted_password_entry)
            return account_name_ok and password_ok
        else:
            host_name_ok = self.check_local_landscape_host_name_entry()
            password_ok = self.check_entry(self.local_password_entry)
            return host_name_ok and password_ok

    @property
    def NO_SERVICE_TEXT(self):
        return _("None")

    @property
    def HOSTED_SERVICE_TEXT(self):
        return _("Landscape - hosted by Canonical")

    @property
    def LOCAL_SERVICE_TEXT(self):
        return _("Landscape - dedicated server")

    @property
    def REGISTER_BUTTON_TEXT(self):
        return _("Register")

    @property
    def DISABLE_BUTTON_TEXT(self):
        return _("Disable")

    @property
    def INVALID_HOST_NAME_MESSAGE(self):
        return _("Invalid host name.")

    @property
    def UNICODE_IN_ENTRY_MESSAGE(self):
        return _("Only ASCII characters are allowed.")

    def _set_use_type_combobox_from_controller(self):
        """
        Load the persisted L{management_type} from the controller and set the
        combobox appropriately.

        Note that Gtk makes us jump through some hoops by having it's own model
        level to deal with here.  The conversion between paths and iters makes
        more sense if you understand that treeviews use the same model.
        """
        list_iter = self.liststore.get_iter_first()
        while (self.liststore.get(list_iter, 0)[0] !=
               self.controller.management_type):
            list_iter = self.liststore.iter_next(list_iter)
        path = self.liststore.get_path(list_iter)
        [index] = path.get_indices()
        self.use_type_combobox.set_active(index)

    def _set_hosted_values_from_controller(self):
        self.hosted_account_name_entry.set_text(
            self.controller.hosted_account_name)
        self.hosted_password_entry.set_text(self.controller.hosted_password)

    def _set_local_values_from_controller(self):
        self.local_landscape_host_entry.set_text(
            self.controller.local_landscape_host)
        self.local_password_entry.set_text(self.controller.local_password)

    def load_data(self):
        self._initialised = False
        self.controller.load()
        self._set_hosted_values_from_controller()
        self._set_local_values_from_controller()
        self._set_use_type_combobox_from_controller()
        self._initialised = True

    def make_liststore(self):
        """
        Construct the correct L{Gtk.ListStore} to drive the L{Gtk.ComboBox} for
        use-type.  This a table of:

           * Management type (key)
           * Text to display in the combobox
           * L{Gtk.Frame} to load when this item is selected.
        """
        liststore = Gtk.ListStore(GObject.TYPE_PYOBJECT,
                                  GObject.TYPE_STRING,
                                  GObject.TYPE_PYOBJECT)
        self.active_widget = None
        liststore.append([NOT_MANAGED, self.NO_SERVICE_TEXT,
                          self._builder.get_object("no-service-frame")])
        liststore.append([CANONICAL_MANAGED, self.HOSTED_SERVICE_TEXT,
                          self._builder.get_object("hosted-service-frame")])
        liststore.append([LOCAL_MANAGED, self.LOCAL_SERVICE_TEXT,
                          self._builder.get_object("local-service-frame")])
        return liststore

    def link_hosted_service_widgets(self):
        self.hosted_account_name_entry = self._builder.get_object(
            "hosted-account-name-entry")
        self.hosted_account_name_entry.connect(
            "changed", self.on_changed_event, "hosted_account_name")

        self.hosted_password_entry = self._builder.get_object(
            "hosted-password-entry")
        self.hosted_password_entry.connect(
            "changed", self.on_changed_event, "hosted_password")

    def link_local_service_widgets(self):
        self.local_landscape_host_entry = self._builder.get_object(
            "local-landscape-host-entry")
        self.local_landscape_host_entry.connect(
            "changed", self.on_changed_event, "local_landscape_host")

        self.local_password_entry = self._builder.get_object(
            "local-password-entry")
        self.local_password_entry.connect(
            "changed", self.on_changed_event, "local_password")

    def link_use_type_combobox(self, liststore):
        self.use_type_combobox = self._builder.get_object("use-type-combobox")
        self.use_type_combobox.connect("changed", self.on_combo_changed)
        self.use_type_combobox.set_model(liststore)
        cell = Gtk.CellRendererText()
        self.use_type_combobox.pack_start(cell, True)
        self.use_type_combobox.add_attribute(cell, 'text', 1)

    def cancel_response(self, widget):
        self.response(Gtk.ResponseType.CANCEL)

    def register_response(self, widget):
        if self.validity_check():
            self.response(Gtk.ResponseType.OK)
        else:
            error_text = []
            if self.UNICODE_IN_ENTRY in self._validation_errors:
                error_text.append(self.UNICODE_IN_ENTRY_MESSAGE)
            if self.INVALID_HOST_NAME in self._validation_errors:
                error_text.append(self.INVALID_HOST_NAME_MESSAGE)
            self.info_message.set_text("\n".join(error_text))
            self._info_bar_container.show()

    def set_button_text(self, management_type):
        [alignment] = self.register_button.get_children()
        [hbox] = alignment.get_children()
        [image, label] = hbox.get_children()
        if management_type == NOT_MANAGED:
            label.set_text(self.DISABLE_BUTTON_TEXT)
        else:
            label.set_text(self.REGISTER_BUTTON_TEXT)

    def setup_buttons(self):
        self.revert_button = Gtk.Button(stock=Gtk.STOCK_REVERT_TO_SAVED)
        self.action_area.pack_start(self.revert_button, True, True, 0)
        self.revert_button.connect("clicked", self.revert)
        self.revert_button.show()
        self.cancel_button = Gtk.Button(stock=Gtk.STOCK_CANCEL)
        self.action_area.pack_start(self.cancel_button, True, True, 0)
        self.cancel_button.show()
        self.cancel_button.connect("clicked", self.cancel_response)
        self.register_button = Gtk.Button(stock=Gtk.STOCK_OK)
        self.action_area.pack_start(self.register_button, True, True, 0)
        self.register_button.show()
        self.register_button.connect("clicked", self.register_response)

    def dismiss_infobar(self, widget):
        self._info_bar_container.hide()
        for entry in self._errored_entries:
            entry.set_icon_from_stock(Gtk.EntryIconPosition.PRIMARY, None)
        self._errored_entries = []

    def setup_info_bar(self):
        labels_size_group = self._builder.get_object("labels-sizegroup")
        entries_size_group = self._builder.get_object("entries-sizegroup")
        labels_size_group.set_ignore_hidden(False)
        entries_size_group.set_ignore_hidden(False)
        self._info_bar_container = Gtk.HBox()
        self._info_bar_container.set_spacing(12)
        info_bar = Gtk.InfoBar()
        entries_size_group.add_widget(info_bar)
        info_bar.show()
        empty_label = Gtk.Label()
        labels_size_group.add_widget(empty_label)
        empty_label.show()
        self._info_bar_container.pack_start(empty_label, expand=False,
                                            fill=False, padding=0)
        self._info_bar_container.pack_start(info_bar, expand=False, fill=False,
                                            padding=0)
        content_area = info_bar.get_content_area()
        hbox = Gtk.HBox()
        self.info_message = Gtk.Label()
        self.info_message.set_alignment(0, 0.5)
        self.info_message.show()
        hbox.pack_start(self.info_message, expand=True, fill=True, padding=6)
        ok_button = Gtk.Button("Dismiss")
        ok_button.connect("clicked", self.dismiss_infobar)
        ok_button.show()
        hbox.pack_start(ok_button, expand=True, fill=True, padding=0)
        hbox.show()
        content_area.pack_start(hbox, expand=True, fill=True, padding=0)

    def setup_ui(self):
        self._builder = Gtk.Builder()
        self._builder.set_translation_domain("landscape-client")
        self._builder.add_from_file(
            os.path.join(
                os.path.dirname(__file__), "ui", self.GLADE_FILE))
        content_area = self.get_content_area()
        content_area.set_spacing(12)
        self.set_border_width(12)
        self.setup_info_bar()
        self._vbox = self._builder.get_object("toplevel-vbox")
        self._vbox.unparent()
        content_area.pack_start(self._vbox, expand=True, fill=True,
                                padding=12)
        self._vbox.pack_start(self._info_bar_container, expand=False,
                              fill=False, padding=0)
        self.liststore = self.make_liststore()
        self.link_use_type_combobox(self.liststore)
        self.link_hosted_service_widgets()
        self.link_local_service_widgets()
        self.setup_buttons()

    def on_combo_changed(self, combobox):
        list_iter = self.liststore.get_iter(combobox.get_active())
        if not self.active_widget is None:
            self._vbox.remove(self.active_widget)
        [management_type] = self.liststore.get(list_iter, 0)
        self.set_button_text(management_type)
        if self._initialised:
            self.controller.management_type = management_type
            self.controller.modify()
        [self.active_widget] = self.liststore.get(list_iter, 2)
        self.active_widget.unparent()
        self._vbox.add(self.active_widget)

    def on_changed_event(self, widget, attribute):
        setattr(self.controller, attribute, widget.get_text())
        self.controller.modify()

    def quit(self, *args):
        self.destroy()

    def revert(self, button):
        self.controller.revert()
        self.load_data()
        # One extra revert to reset after loading data
        self.controller.revert()



"""Provide backward compatible access to hashlib functions."""


try:
    _hashlib = __import__("hashlib")
except ImportError:
    from md5 import md5
    from sha import sha as sha1
else:
    md5 = _hashlib.md5
    sha1 = _hashlib.sha1


__all__ = ["md5", "sha1"]



import os
import pwd
import logging

from twisted.internet.error import ProcessDone
from twisted.python.failure import Failure

from landscape import SERVER_API

from landscape.manager.customgraph import CustomGraphPlugin
from landscape.manager.store import ManagerStore

from landscape.tests.helpers import (
    LandscapeTest, ManagerHelper, StubProcessFactory, DummyProcess)
from landscape.tests.mocker import ANY


class CustomGraphManagerTests(LandscapeTest):
    helpers = [ManagerHelper]

    def setUp(self):
        super(CustomGraphManagerTests, self).setUp()
        self.store = ManagerStore(":memory:")
        self.manager.store = self.store
        self.broker_service.message_store.set_accepted_types(
            ["custom-graph"])
        self.data_path = self.makeDir()
        self.manager.config.data_path = self.data_path
        os.makedirs(os.path.join(self.data_path, "custom-graph-scripts"))
        self.manager.config.script_users = "ALL"
        self.graph_manager = CustomGraphPlugin(
            create_time=range(1500, 0, -300).pop)
        self.manager.add(self.graph_manager)

    def _exit_process_protocol(self, protocol, stdout):
        protocol.childDataReceived(1, stdout)
        for fd in (0, 1, 2):
            protocol.childConnectionLost(fd)
        protocol.processEnded(Failure(ProcessDone(0)))

    def test_add_graph(self):
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo hi!",
                     "username": username,
                     "graph-id": 123})

        self.assertEqual(
            self.store.get_graphs(),
            [(123,
              os.path.join(self.data_path, "custom-graph-scripts",
                           "graph-123"),
              username)])

    def test_add_graph_unknown_user(self):
        """
        Attempting to add a graph with an unknown user should not result in an
        error, instead a message should be logged, the error will be picked up
        when the graph executes.
        """
        mock_getpwnam = self.mocker.replace("pwd.getpwnam", passthrough=False)
        mock_getpwnam("foo")
        self.mocker.throw(KeyError("foo"))
        self.mocker.replay()
        error_message = "Attempt to add graph with unknown user foo"
        self.log_helper.ignore_errors(error_message)
        self.logger.setLevel(logging.ERROR)

        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo hi!",
                     "username": "foo",
                     "graph-id": 123})
        graph = self.store.get_graph(123)
        self.assertEqual(graph[0], 123)
        self.assertEqual(graph[2], u"foo")
        self.assertTrue(error_message in self.logfile.getvalue())

    def test_add_graph_for_user(self):
        mock_chown = self.mocker.replace("os.chown", passthrough=False)
        mock_chown(ANY, 1234, 5678)

        mock_chmod = self.mocker.replace("os.chmod", passthrough=False)
        mock_chmod(ANY, 0700)

        mock_getpwnam = self.mocker.replace("pwd.getpwnam", passthrough=False)

        class pwnam(object):
            pw_uid = 1234
            pw_gid = 5678
            pw_dir = self.makeFile()

        self.expect(mock_getpwnam("bar")).result(pwnam)
        self.mocker.replay()
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo hi!",
                     "username": "bar",
                     "graph-id": 123})
        self.assertEqual(
            self.store.get_graphs(),
            [(123, os.path.join(self.data_path, "custom-graph-scripts",
                                "graph-123"),
                   "bar")])

    def test_remove_unknown_graph(self):
        self.manager.dispatch_message(
            {"type": "custom-graph-remove",
                     "graph-id": 123})

    def test_remove_graph(self):
        filename = self.makeFile()
        tempfile = file(filename, "w")
        tempfile.write("foo")
        tempfile.close()
        self.store.add_graph(123, filename, u"user")
        self.manager.dispatch_message(
            {"type": "custom-graph-remove",
                     "graph-id": 123})
        self.assertFalse(os.path.exists(filename))

    def test_run(self):
        filename = self.makeFile()
        tempfile = file(filename, "w")
        tempfile.write("#!/bin/sh\necho 1")
        tempfile.close()
        os.chmod(filename, 0777)
        self.store.add_graph(123, filename, None)

        def check(ignore):
            self.graph_manager.exchange()
            script_hash = "483f2304b49063680c75e3c9e09cf6d0"
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data": {123: {"error": u"",
                                 "values": [(300, 1.0)],
                                 "script-hash": script_hash}},
                  "type": "custom-graph"}])
        return self.graph_manager.run().addCallback(check)

    def test_run_multiple(self):
        filename = self.makeFile()
        tempfile = file(filename, "w")
        tempfile.write("#!/bin/sh\necho 1")
        tempfile.close()
        os.chmod(filename, 0777)
        self.store.add_graph(123, filename, None)

        filename = self.makeFile()
        tempfile = file(filename, "w")
        tempfile.write("#!/bin/sh\necho 2")
        tempfile.close()
        os.chmod(filename, 0777)
        self.store.add_graph(124, filename, None)

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data":
                      {123: {"error": u"",
                             "values": [(300, 1.0)],
                             "script-hash": "483f2304b49063680c75e3c9e09cf6d0"
                            },
                       124: {"error": u"",
                             "values": [(300, 2.0)],
                             "script-hash": "73a74b1530b2256db7edacb9b9cc385e"
                            }
                      },
                  "type": "custom-graph"}])
        return self.graph_manager.run().addCallback(check)

    def test_run_with_nonzero_exit_code(self):
        filename = self.makeFile()
        tempfile = file(filename, "w")
        tempfile.write("#!/bin/sh\nexit 1")
        tempfile.close()
        os.chmod(filename, 0777)
        self.store.add_graph(123, filename, None)

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data":
                      {123: {"error": u" (process exited with code 1)",
                             "values": [],
                             "script-hash": "eaca3ba1a3bf1948876eba320148c5e9"
                            }
                      },
                  "type": "custom-graph"}])
        return self.graph_manager.run().addCallback(check)

    def test_run_cast_result_error(self):
        filename = self.makeFile("some_content")
        self.store.add_graph(123, filename, None)
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 1)
        spawn = factory.spawns[0]
        self.assertEqual(spawn[1], filename)

        self._exit_process_protocol(spawn[0], "foobar")

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data":
                      {123: {"error":
                             u"InvalidFormatError: Failed to convert to "
                              "number: 'foobar'",
                             "values": [], "script-hash":
                                 "baab6c16d9143523b7865d46896e4596"}},
                  "type": "custom-graph"}])
        return result.addCallback(check)

    def test_run_no_output_error(self):
        filename = self.makeFile("some_content")
        self.store.add_graph(123, filename, None)
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 1)
        spawn = factory.spawns[0]
        self.assertEqual(spawn[1], filename)

        self._exit_process_protocol(spawn[0], "")

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data":
                      {123: {"error": u"NoOutputError: Script did not output "
                                       "any value",
                             "values": [], "script-hash":
                                 "baab6c16d9143523b7865d46896e4596"}},
                  "type": "custom-graph"}])
        return result.addCallback(check)

    def test_run_no_output_error_with_other_result(self):
        filename1 = self.makeFile("some_content")
        self.store.add_graph(123, filename1, None)
        filename2 = self.makeFile("some_content")
        self.store.add_graph(124, filename2, None)
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 2)
        spawn = factory.spawns[0]
        self._exit_process_protocol(spawn[0], "")
        spawn = factory.spawns[1]
        self._exit_process_protocol(spawn[0], "0.5")

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data":
                      {123: {"error": u"NoOutputError: Script did not output "
                                       "any value",
                             "script-hash": "baab6c16d9143523b7865d46896e4596",
                             "values": []},
                       124: {"error": u"",
                             "script-hash": "baab6c16d9143523b7865d46896e4596",
                             "values": [(300, 0.5)]}},
                  "type": "custom-graph"}])
        return result.addCallback(check)

    def test_multiple_errors(self):
        filename1 = self.makeFile("some_content")
        self.store.add_graph(123, filename1, None)
        filename2 = self.makeFile("some_content")
        self.store.add_graph(124, filename2, None)
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 2)
        spawn = factory.spawns[0]
        self._exit_process_protocol(spawn[0], "foo")
        spawn = factory.spawns[1]
        self._exit_process_protocol(spawn[0], "")

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data":
                      {123: {"error": u"InvalidFormatError: Failed to convert "
                                       "to number: 'foo'",
                             "script-hash": "baab6c16d9143523b7865d46896e4596",
                             "values": []},
                       124: {"error": u"NoOutputError: Script did not output "
                                       "any value",
                             "script-hash": "baab6c16d9143523b7865d46896e4596",
                             "values": []}},
                  "type": "custom-graph"}])
        return result.addCallback(check)

    def test_run_user(self):
        filename = self.makeFile("some content")
        self.store.add_graph(123, filename, "bar")
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory

        mock_getpwnam = self.mocker.replace("pwd.getpwnam", passthrough=False)

        class pwnam(object):
            pw_uid = 1234
            pw_gid = 5678
            pw_dir = self.makeFile()

        self.expect(mock_getpwnam("bar")).result(pwnam)
        self.mocker.replay()

        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 1)
        spawn = factory.spawns[0]
        self.assertEqual(spawn[1], filename)
        self.assertEqual(spawn[2], ())
        self.assertEqual(spawn[3], {})
        self.assertEqual(spawn[4], "/")
        self.assertEqual(spawn[5], 1234)
        self.assertEqual(spawn[6], 5678)

        self._exit_process_protocol(spawn[0], "spam")

        return result

    def test_run_dissallowed_user(self):
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        self.manager.config.script_users = "foo"

        self.store.add_graph(123, "filename", username)
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 0)

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data": {123:
                      {"error":
                       u"ProhibitedUserError: Custom graph cannot be run as "
                        "user %s" % (username,),
                       "script-hash": "",
                       "values": []}},
                  "type": "custom-graph"}])

        return result.addCallback(check)

    def test_run_unknown_user(self):
        mock_getpwnam = self.mocker.replace("pwd.getpwnam", passthrough=False)
        mock_getpwnam("foo")
        self.mocker.throw(KeyError("foo"))
        self.mocker.replay()

        self.manager.config.script_users = "foo"

        self.store.add_graph(123, "filename", "foo")
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 0)

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data": {123:
                      {"error": u"UnknownUserError: Unknown user 'foo'",
                       "script-hash": "",
                       "values": []}},
                  "type": "custom-graph"}])

        return result.addCallback(check)

    def test_run_timeout(self):
        filename = self.makeFile("some content")
        self.store.add_graph(123, filename, None)
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 1)
        spawn = factory.spawns[0]
        protocol = spawn[0]
        protocol.makeConnection(DummyProcess())
        self.assertEqual(spawn[1], filename)

        self.manager.reactor.advance(110)
        protocol.processEnded(Failure(ProcessDone(0)))

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data": {123: {"error":
                                    u"Process exceeded the 10 seconds limit",
                                "script-hash":
                                    "9893532233caff98cd083a116b013c0b",
                                "values": []}},
                  "type": "custom-graph"}])

        return result.addCallback(check)

    def test_run_removed_file(self):
        """
        If run is called on a script file that has been removed, it doesn't try
        to run it, but report it with an empty hash value.
        """
        self.store.add_graph(123, "/nonexistent", None)
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 0)

        self.graph_manager.exchange()
        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(),
            [{"data": {123: {"error": u"", "script-hash": "", "values": []}},
              "type": "custom-graph"}])

    def test_send_message_add_stored_graph(self):
        """
        C{send_message} send the graph with no data, to notify the server of
        the existence of the script, even if the script hasn't been run yet.
        """
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo hi!",
                     "username": username,
                     "graph-id": 123})
        self.graph_manager.exchange()
        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(),
            [{"api": SERVER_API,
              "data": {123: {"error": u"",
                             "script-hash": "e00a2f44dbc7b6710ce32af2348aec9b",
                             "values": []}},
              "timestamp": 0,
              "type": "custom-graph"}])

    def test_send_message_check_not_present_graph(self):
        """C{send_message} checks the presence of the custom-graph script."""
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo hi!",
                     "username": username,
                     "graph-id": 123})
        filename = self.store.get_graph(123)[1]
        os.unlink(filename)
        self.graph_manager.exchange()
        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(),
            [{"api": SERVER_API,
              "data": {},
              "timestamp": 0,
              "type": "custom-graph"}])

    def test_send_message_dont_rehash(self):
        """
        C{send_message} uses hash already stored if still no data has been
        found.
        """
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo hi!",
                     "username": username,
                     "graph-id": 123})
        self.graph_manager.exchange()
        self.graph_manager._get_script_hash = lambda x: 1 / 0
        self.graph_manager.do_send = True
        self.graph_manager.exchange()
        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(),
            [{"api": SERVER_API,
              "data": {123: {"error": u"",
                             "script-hash": "e00a2f44dbc7b6710ce32af2348aec9b",
                             "values": []}},
              "timestamp": 0,
              "type": "custom-graph"},
             {"api": SERVER_API,
              "data": {123: {"error": u"",
                             "script-hash": "e00a2f44dbc7b6710ce32af2348aec9b",
                             "values": []}},
              "timestamp": 0,
              "type": "custom-graph"}])

    def test_send_message_rehash_if_necessary(self):
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo hi!",
                     "username": username,
                     "graph-id": 123})
        self.graph_manager.exchange()
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo bye!",
                     "username": username,
                     "graph-id": 123})
        self.graph_manager.do_send = True
        self.graph_manager.exchange()
        self.assertMessages(
            self.broker_service.message_store.get_pending_messages(),
            [{"api": SERVER_API,
              "data": {123: {"error": u"",
                             "script-hash": "e00a2f44dbc7b6710ce32af2348aec9b",
                             "values": []}},
              "timestamp": 0,
              "type": "custom-graph"},
             {"api": SERVER_API,
              "data": {123: {"error": u"",
                             "script-hash": "d483816dc0fbb51ede42502a709b0e2a",
                             "values": []}},
              "timestamp": 0,
              "type": "custom-graph"}])

    def test_run_with_script_updated(self):
        """
        If a script is updated while a data point is being retrieved, the data
        point is discarded and no value is sent, but the new script is
        mentioned.
        """
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
            "code": "echo 1.0",
                     "username": username,
                     "graph-id": 123})

        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 1)
        spawn = factory.spawns[0]

        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo 2.0",
                     "username": username,
                     "graph-id": 123})

        self._exit_process_protocol(spawn[0], "1.0")

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"api": SERVER_API,
                  "data": {123: {"error": u"",
                                 "script-hash":
                                    "991e15a81929c79fe1d243b2afd99c62",
                                 "values": []}},
                  "timestamp": 0,
                  "type": "custom-graph"}])

        return result.addCallback(check)

    def test_run_with_script_removed(self):
        """
        If a script is removed while a data point is being retrieved, the data
        point is discarded and no data is sent at all.
        """
        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo 1.0",
                     "username": username,
                     "graph-id": 123})

        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 1)
        spawn = factory.spawns[0]

        self.manager.dispatch_message(
            {"type": "custom-graph-remove",
                     "graph-id": 123})

        self._exit_process_protocol(spawn[0], "1.0")

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"api": SERVER_API, "data": {}, "timestamp": 0, "type":
                  "custom-graph"}])
        return result.addCallback(check)

    def test_run_not_accepted_types(self):
        """
        If "custom-graph" is not an accepted message-type anymore,
        C{CustomGraphPlugin.run} shouldn't even run the graph scripts.
        """
        self.broker_service.message_store.set_accepted_types([])

        uid = os.getuid()
        info = pwd.getpwuid(uid)
        username = info.pw_name
        self.manager.dispatch_message(
            {"type": "custom-graph-add",
                     "interpreter": "/bin/sh",
                     "code": "echo 1.0",
                     "username": username,
                     "graph-id": 123})

        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 0)

        return result.addCallback(self.assertIdentical, None)

    def test_run_without_graph(self):
        """
        If no graph is available, C{CustomGraphPlugin.run} doesn't even call
        C{call_if_accepted} on the broker and return immediately an empty list
        of results.
        """
        self.graph_manager.registry.broker.call_if_accepted = (
            lambda *args: 1 / 0)
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 0)

        return result.addCallback(self.assertEqual, [])

    def test_run_unknown_user_with_unicode(self):
        """
        Using a non-existent user containing unicode characters fails with the
        appropriate error message.
        """
        username = u"non-existent-f\N{LATIN SMALL LETTER E WITH ACUTE}e"

        self.manager.config.script_users = "ALL"

        filename = self.makeFile("some content")
        self.store.add_graph(123, filename, username)
        factory = StubProcessFactory()
        self.graph_manager.process_factory = factory
        result = self.graph_manager.run()

        self.assertEqual(len(factory.spawns), 0)

        def check(ignore):
            self.graph_manager.exchange()
            self.assertMessages(
                self.broker_service.message_store.get_pending_messages(),
                [{"data": {123:
                      {"error":
                           u"UnknownUserError: Unknown user '%s'" % username,
                       "script-hash": "9893532233caff98cd083a116b013c0b",
                       "values": []}},
                  "type": "custom-graph"}])

        return result.addCallback(check)



import json
import logging
import os.path

from landscape.lib.fs import read_file


def get_juju_info(config):
    """
    Returns the Juju info or C{None} if the path referenced from
    L{config} is not a file or that file isn't valid JSON.
    """
    juju_filename = config.juju_filename
    if not os.path.isfile(juju_filename):
        return None
    json_contents = read_file(juju_filename)
    try:
        juju_info = json.loads(json_contents)
    except Exception:
        logging.exception(
            "Error attempting to read JSON from %s" % juju_filename)
        return None
    else:
        if "api-addresses" in juju_info:
            juju_info["api-addresses"] = juju_info["api-addresses"].split()
        return juju_info



from landscape.diff import diff
from landscape.tests.helpers import LandscapeTest


class DiffTest(LandscapeTest):

    def test_empty(self):
        self.assertEqual(diff({}, {}), ({}, {}, {}))

    def test_identical(self):
        data = {"str": "wubble", "strlist": ["foo", "bar"]}
        self.assertEqual(diff(data, data), ({}, {}, {}))

    def test_create(self):
        old = {}
        new = {"str": "wubble"}
        self.assertEqual(diff(old, new), ({"str": "wubble"}, {}, {}))

    def test_update(self):
        old = {"str": "wubble"}
        new = {"str": "ooga"}
        self.assertEqual(diff(old, new), ({}, {"str": "ooga"}, {}))

    def test_delete(self):
        old = {"str": "wubble"}
        new = {}
        self.assertEqual(diff(old, new), ({}, {}, {"str": "wubble"}))

    def test_complex(self):
        old = {"str": "wubble", "int": 10}
        new = {"strlist": ["foo", "bar"], "int": 25}
        self.assertEqual(diff(old, new), ({"strlist": ["foo", "bar"]},
                                          {"int": 25},
                                          {"str": "wubble"}))



"""Manage outgoing and incoming messages when communicating with the server.

The protocol to communicate between the client and the server has been designed
to be very robust so that messages are not lost. In addition it is (vaguely)
symmetric, as the client and server need to send messages both ways.

Client->Server Payload
======================

All message payloads are bpickled with L{landscape.lib.bpickle.dumps}. Client
to server payloads are C{dict}s of the form::

  {'server-api': SERVER_API_VERSION,
   'client-api': CLIENT_API_VERSION,
   'sequence': SEQUENCE_NUMBER,
   'accepted-types': SERVER_ACCEPTED_TYPES_DIGEST,
   'messages': MESSAGES,
   'total-messages': TOTAL_COUNT_OF_PENDING_MESSAGES,
   'next-expected-sequence': EXPECTED_SEQUENCE_NUMBER,
   'client-accepted-types': CLIENT_ACCEPTED_TYPES (optional)}

The values have the following semantics:

  - C{SERVER_API_VERSION}: The API version that is required on the server
    in order to process the messages in this payload (the schema and semantics
    of message types are usually different for different API versions).

  - C{CLIENT_API_VERSION}: The API version of the client, hinting the server
    about the schema and semantics of the messages types accepted by the client
    (see below).

  - C{SEQUENCE_NUMBER}: A monotonically increasing nonnegative integer. The
    meaning of this is described below.

  - C{SERVER_ACCEPTED_TYPES_DIGEST}: A hash of the message types that the
    client thinks are currently accepted by the server. The server can use it
    to know whether to send the client a new up-to-date list of accepted
    message types.

  - C{MESSAGES}: A python list of messages, described below.

  - C{TOTAL_COUNT_OF_PENDING_MESSAGES}: The total number of messages in the
    client outgoing queue. This is includes the number of messages being sent
    in this payload, plus any other messages still pending and not included
    here.

  - C{EXPECTED_SEQUENCE_NUMBER}: The sequence number which the client expects
    the next message sent from the server to have.

  - C{CLIENT_ACCEPTED_TYPES}: Optionally, a list of message types that the
    client accepts. The server is supposed to send the client only messages of
    this type. It will be inlcuded in the payload only if the hash that the
    server sends us is out-of-date. This behavior is simmetric with respect to
    the C{SERVER_ACCEPTED_TYPES_DIGEST} field described above.

Server->Client Payload
======================

The payloads that the server sends to not-yet-registered clients (i.e. clients
that don't provide a secure ID associated with a computer) are C{dict}s of the
form::

  {'server-uuid': SERVER_UUID,
   'messages': MESSAGES}

where:

  - C{SERVER_UUID}: A string identifying the particular Landscape server the
    client is talking to.

  - C{MESSAGES}: A python list of messages, described below.

Additionally, payloads to registered clients will include these fields::

  {'next-expected-sequence': EXPECTED_SEQUENCE_NUMBER,
   'client-accepted-types-hash': CLIENT_ACCEPTED_TYPES_DIGEST,

where:

  - C{EXPECTED_SEQUENCE_NUMBER}: The sequence number which the server expects
    the next message sent from the client to have.

  - C{CLIENT_ACCEPTED_TYPES_DIGEST}: A hash of the message types that the
    server thinks are currently accepted by the client. The client can use it
    to know whether to send to the server an up-to-date list the message types
    it now accepts (see CLIENT_ACCEPTED_TYPES in the client->server payload).

Individual Messages
===================

A message is a C{dict} with required and optional keys. Messages are packed
into Python lists and set as the value of the 'messages' key in the payload.

The C{dict} of a single message is of the form::

  {'type': MESSAGE_TYPE,
   ...}

where:

  - C{MESSAGE_TYPE}: A simple string, which lets the server decide what handler
    to dispatch the message to, also considering the SERVER_API_VERSION value.

  - C{...}: Other entries specific to the type of message.

Message Sequencing
==================

A message numbering system is built in to the protocol to ensure robustness of
client/server communication. The way this works is not totally symmetrical, as
the client must connect to the server via HTTP, but the ordering that things
happen in over the course of many connections remains the same (see also
L{landscape.broker.store} for more concrete examples):

  - Receiver tells Sender which sequence number it expects the next batch of
    messages to start with.

  - Sender gives some messages to Receiver, specifying the sequence number of
    the first message. If the expected and actual sequence numbers are out of
    synch, Sender resynchronizes in a certain way.

The client and server must play the part of *both* of these roles on every
interaction, but it simplifies things to talk about them in terms of a single
role at a time.

When the client connects to the server, it does the following things acting
in the role of Sender (which is by far its more burdened role):

  - Send a payload containing messages and a sequence number. The sequence
    number should be the same number that the server gave as
    next-expected-sequence in the prior connection, or 0 if there was no
    previous connection.

  - Get back a next-expected-sequence from the server. If that value is is not
    len(messages) + previous-next-expected, then resynchronize.

It does the following when acting as Receiver:

  - Send a payload containing a next-expected-sequence, which should be the
    sequence number of the first message that the server responds with. This
    value should be previous-next-expected + len(previous_messages).

  - Receive some messages from the server, and process them immediately.

When the server is acting as Sender, it does the following:

  - Wait for a payload with next-expected-sequence from the client.

  - Perhaps resynchronize if next-expected-sequence is unexpected.

  - Respond with a payload of messages to the client. No sequence identifier
    is given for this payload of messages, because it would be redundant with
    data that has already passed over the wire (received from the client)
    during the very same TCP connection.

When the server is acting as a Receiver, it does the following:

  - Wait for a payload with a sequence identifier and a load of messages.
  - Respond with a next-expected-sequence.

There are two interesting exceptional cases which must be handled with
resynchronization:

  1. Messages received with sequence numbers less than the next expected
     sequence number should be discarded, and further messages starting at
     the expected sequence numbers should be processed.

  2. If the sequence number is higher than what the receiver expected, then
     no messages are processed and the receiver responds with the same
     {'next-expected-sequence': N}, so that the sender can resynchronize
     itself.

This implies that the receiver must record the sequence number of the last
successfully processed message, in order for it to respond to the sender
with that number. In addition, the sender must save outbound messages even
after they have been delivered over the transport, until the sender receives
a next-expected-sequence higher than the outbound message. The details of
this logic are described in L{landscape.broker.store}.

Exchange Sequence
=================

Diagram::

  1. BrokerService    -->  MessageExchange               : Start

  2. MessageExchange  -->  MessageExchange               : Schedule exchange

  3. [event]          <--  MessageExchange               : Fire "pre-exchange"

  4. [optional]                                          : Do registration
     (See L{landscape.broker.registration})              : sequence

  5. MessageExchange  -->  MessageStore                  : Request pending
                                                         : messages

  6. MessageExchange  <--  MessageStore                  : return( Messages )

  7. MessageExchange  -->  HTTPTransport                 : Exchange

  8. HTTPTransport    -->  {Server}LandscapeMessageSystem
                                                         : HTTP POST

  9. [Scope: Server]
   |
   |   9.1 LandscapeMessageSystem --> ComputerMessageAPI : run
   |
   |   9.2 ComputerMessageAPI     --> FunctionHandler    : handle
   |
   |   9.3 FunctionHandler        --> Callable           : call
   |       ( See also server code at:
   |             - C{canonical.landscape.message.handlers}
   |             - C{canonical.message.handler.FunctionHandler} )
   |
   |
   |   9.4 [If: the callable raises ConsistencyError]
   |     |
   |     | 9.4.1 ComputerMessageAPI --> Computer         : request
   |     |                                               : Resynchronize
   |     |
   |     | 9.4.2 Computer           --> Computer         : Create
   |     |                                               : ResynchronizeRequest
   |     |                                               : activity
   |     |
   |     --[End If]
   |
   |  9.5 ComputerMessageAPI     --> Computer            : get deliverable
   |                                                     : activities
   |
   |  9.6 ComputerMessageAPI     <-- Computer            : return activities
   |
   |  9.7 [Loop over activities]
   |    |
   |    | 9.7.1 ComputerMessageAPI  --> Activity         : deliver
   |    |
   |    | 9.7.2 Activity            --> MessageStore     : add activity message
   |    |
   |    --[End Loop]
   |
   |  9.8 ComputerMessageAPI     --> MessageStore        : get pending messages
   |
   |  9.9 ComputerMessageAPI     <-- MessageStore        : return messages
   |
   | 9.10 LandscapeMessageSystem <-- ComputerMessageAPI  : return payload
   |                                                     : (See below)
   |
   -- [End Scope]

  10. HTTPTransport    <--  {Server}LandscapeMessageSystem
                                                         : HTTP response
                                                         : with payload

  11. MessageExchange  <--  HTTPTransport                : response

  12. [If: server says it expects a very old message]
   |
   |  12.1 [event]              <-- MessageExchange      : event
   |       (See L{landscape.broker.server})              : "resynchronize-
   |                                                     : clients"
   |
   -- [End if]

  13. [Loop: over messages in payload]
   |
   |  13.1 [event]             <-- MessageExchange       : event
   |                                                     : message (message)
   |
   |  13.2 [Switch: on message type]
   |     |
   |     |- 13.2.1 [Case: message type is "accepted-types"]
   |     |       |
   |     |       | 13.2.1.1 MessageExchange -> MessageStore
   |     |       |                                       : set accepted types
   |     |       |
   |     |       | 13.2.1.2 MessageExchange -> MessageExchange
   |     |       |                                       : schedule urgent
   |     |       |                                       : exchange
   |     |       --[End Case]
   |     |
   |     |- 13.2.2 [Case: message type is "resynchronize"]
   |     |       |
   |     |       | 13.2.2.1 [event]         <- MessageExchange
   |     |       |        (See L{landscape.broker.server})
   |     |       |                                      : event
   |     |       |                                      : "resynchronize-
   |     |       |                                      : clients"
   |     |       |
   |     |       | 13.2.2.2 MessageExchange -> MessageStore
   |     |       |                                      : add "resynchronize"
   |     |       |                                      : message
   |     |       |
   |     |       | 13.2.2.3 MessageExchange -> MessageExchange
   |     |       |                                      : schedule urgent
   |     |       |                                      : exchange
   |     |       |
   |     |       --[End Case]
   |     |
   |     |- 13.2.3 [Case: message type is "set-intervals"]
   |     |       |
   |     |       | 13.2.3.1 MessageExchange -> BrokerConfiguration
   |     |       |                                      : set exchange
   |     |       |                                      : interval
   |     |       |
   |     |       --[End Case]
   |     |
   |     -- [End Switch]
   |
   -- [End Loop]

  14. Schedule exchange

"""
import time
import logging
from landscape.lib.hashlib import md5

from twisted.internet.defer import Deferred, succeed

from landscape.lib.message import got_next_expected, ANCIENT
from landscape.log import format_delta
from landscape import SERVER_API, CLIENT_API


class MessageExchange(object):
    """Schedule and handle message exchanges with the server.

    The L{MessageExchange} is the place where messages are sent to go out
    to the Landscape server. It accumulates messages in its L{MessageStore}
    and periodically delivers them to the server.

    It is also the place where messages coming from the server are handled. For
    each message type the L{MessageExchange} supports setting an handler that
    will be invoked when a message of the that type is received.

    An exchange is performed with an HTTP POST request, whose body contains
    outgoing messages and whose response contains incoming messages.
    """

    def __init__(self, reactor, store, transport, registration_info,
                 exchange_store, config,  max_messages=100):
        """
        @param reactor: The L{LandscapeReactor} used to fire events in response
            to messages received by the server.
        @param store: The L{MessageStore} used to queue outgoing messages.
        @param transport: The L{HTTPTransport} used to deliver messages.
        @param registration_info: The L{Identity} storing our secure ID.
        @param config: The L{BrokerConfiguration} with the `exchange_interval`
            and `urgent_exchange_interval` parameters, respectively holding
            the time interval between subsequent exchanges of non-urgent
            messages, and the time interval between subsequent exchanges
            of urgent messages.
        """
        self._reactor = reactor
        self._message_store = store
        self._transport = transport
        self._registration_info = registration_info
        self._config = config
        self._exchange_interval = config.exchange_interval
        self._urgent_exchange_interval = config.urgent_exchange_interval
        self._max_messages = max_messages
        self._notification_id = None
        self._exchange_id = None
        self._exchanging = False
        self._urgent_exchange = False
        self._client_accepted_types = set()
        self._client_accepted_types_hash = None
        self._message_handlers = {}
        self._exchange_store = exchange_store
        self._stopped = False

        self.register_message("accepted-types", self._handle_accepted_types)
        self.register_message("resynchronize", self._handle_resynchronize)
        self.register_message("set-intervals", self._handle_set_intervals)
        reactor.call_on("resynchronize-clients", self._resynchronize)

    def _message_is_obsolete(self, message):
        """Returns C{True} if message is obsolete.

        A message is considered obsolete if the secure ID changed since it was
        received.
        """
        if 'operation-id' not in message:
            return False

        operation_id = message['operation-id']
        context = self._exchange_store.get_message_context(operation_id)
        if context is None:
            logging.warning(
                "No message context for message with operation-id: %s"
                % operation_id)
            return False

        # Compare the current secure ID with the one that was in effect when
        # the request message was received.
        result = self._registration_info.secure_id != context.secure_id
        context.remove()

        return result

    def send(self, message, urgent=False):
        """Include a message to be sent in an exchange.

        If urgent is True, an exchange with the server will be
        scheduled urgently.

        @param message: Same as in L{MessageStore.add}.
        """
        if self._message_is_obsolete(message):
            logging.info(
                "Response message with operation-id %s was discarded "
                "because the client's secure ID has changed in the meantime"
                % message.get('operation-id'))
            return None

        if "timestamp" not in message:
            message["timestamp"] = int(self._reactor.time())
        message_id = self._message_store.add(message)
        if urgent:
            self.schedule_exchange(urgent=True)
        return message_id

    def start(self):
        """Start scheduling exchanges. The first one will be urgent."""
        self.schedule_exchange(urgent=True)

    def stop(self):
        """Stop scheduling exchanges."""
        if self._exchange_id is not None:
            # Cancel the next scheduled exchange
            self._reactor.cancel_call(self._exchange_id)
            self._exchange_id = None
        if self._notification_id is not None:
            # Cancel the next scheduled notification of an impending exchange
            self._reactor.cancel_call(self._notification_id)
            self._notification_id = None
        self._stopped = True

    def _handle_accepted_types(self, message):
        """
        When the server updates us about the types of message it
        accepts, update our message store.

        If this makes existing held messages available for sending,
        urgently exchange messages.

        If new types are made available or old types are dropped a
        C{("message-type-acceptance-changed", type, bool)} reactor
        event will be fired.
        """
        old_types = set(self._message_store.get_accepted_types())
        new_types = set(message["types"])
        diff = get_accepted_types_diff(old_types, new_types)
        self._message_store.set_accepted_types(new_types)
        logging.info("Accepted types changed: %s", diff)
        if self._message_store.get_pending_messages(1):
            self.schedule_exchange(urgent=True)
        for type in old_types - new_types:
            self._reactor.fire("message-type-acceptance-changed", type, False)
        for type in new_types - old_types:
            self._reactor.fire("message-type-acceptance-changed", type, True)

    def _handle_resynchronize(self, message):
        opid = message["operation-id"]
        scopes = message.get("scopes")
        self.send({"type": "resynchronize", "operation-id": opid})
        self._reactor.fire("resynchronize-clients", scopes=scopes)

    def _resynchronize(self, scopes=None):
        # When re-synchronisation occurs we don't want any previous messages
        # being sent to the server, dropping the existing session_ids means
        # that messages sent with those IDs will be dropped by the broker.
        self._message_store.drop_session_ids(scopes)
        self.schedule_exchange(urgent=True)

    def _handle_set_intervals(self, message):
        if "exchange" in message:
            self._config.exchange_interval = message["exchange"]
            logging.info("Exchange interval set to %d seconds." %
                         self._config.exchange_interval)
        if "urgent-exchange" in message:
            self._config.urgent_exchange_interval = message["urgent-exchange"]
            logging.info("Urgent exchange interval set to %d seconds." %
                         self._config.urgent_exchange_interval)
        self._config.write()

    def exchange(self):
        """Send pending messages to the server and process responses.

        A C{pre-exchange} reactor event will be emitted just before the
        actual exchange takes place.

        An C{exchange-done} or C{exchange-failed} reactor event will be
        emitted after a successful or failed exchange.

        @return: A L{Deferred} that is fired when exchange has completed.
        """
        if self._exchanging:
            return succeed(None)

        self._exchanging = True

        self._reactor.fire("pre-exchange")

        payload = self._make_payload()

        start_time = time.time()
        if self._urgent_exchange:
            logging.info("Starting urgent message exchange with %s."
                         % self._transport.get_url())
        else:
            logging.info("Starting message exchange with %s."
                         % self._transport.get_url())

        deferred = Deferred()

        def exchange_completed():
            self.schedule_exchange(force=True)
            self._reactor.fire("exchange-done")
            logging.info("Message exchange completed in %s.",
                         format_delta(time.time() - start_time))
            deferred.callback(None)

        def handle_result(result):
            self._exchanging = False
            if result:
                if self._urgent_exchange:
                    logging.info("Switching to normal exchange mode.")
                    self._urgent_exchange = False
                self._handle_result(payload, result)
                self._message_store.record_success(int(self._reactor.time()))
            else:
                self._reactor.fire("exchange-failed")
                logging.info("Message exchange failed.")
            exchange_completed()

        def handle_failure(failure_type, failure_value, failure_tb):
            self._exchanging = False
            self._reactor.fire("exchange-failed")
            self._message_store.record_failure(int(self._reactor.time()))
            logging.info("Message exchange failed.")
            exchange_completed()

        self._reactor.call_in_thread(handle_result, handle_failure,
                                     self._transport.exchange, payload,
                                     self._registration_info.secure_id,
                                     self._get_exchange_token(),
                                     payload.get("server-api"))
        return deferred

    def is_urgent(self):
        """Return a bool showing whether there is an urgent exchange scheduled.
        """
        return self._urgent_exchange

    def schedule_exchange(self, urgent=False, force=False):
        """Schedule an exchange to happen.

        The exchange will occur after some time based on whether C{urgent} is
        True. An C{impending-exchange} reactor event will be emitted
        approximately 10 seconds before the exchange is started.

        @param urgent: If true, ensure an exchange happens within the
            urgent interval.  This will reschedule the exchange if necessary.
            If another urgent exchange is already scheduled, nothing happens.
        @param force: If true, an exchange will necessarily be scheduled,
            even if it was already scheduled before.
        """
        if self._stopped:
            return
        # The 'not self._exchanging' check below is currently untested.
        # It's a bit tricky to test as it is preventing rehooking 'exchange'
        # while there's a background thread doing the exchange itself.
        if (not self._exchanging and
            (force or self._exchange_id is None or
             urgent and not self._urgent_exchange)):
            if urgent:
                self._urgent_exchange = True
            if self._exchange_id:
                self._reactor.cancel_call(self._exchange_id)

            if self._urgent_exchange:
                interval = self._config.urgent_exchange_interval
            else:
                interval = self._config.exchange_interval

            if self._notification_id is not None:
                self._reactor.cancel_call(self._notification_id)
            notification_interval = interval - 10
            self._notification_id = self._reactor.call_later(
                notification_interval, self._notify_impending_exchange)

            self._exchange_id = self._reactor.call_later(
                interval, self.exchange)

    def _get_exchange_token(self):
        """Get the token given us by the server at the last exchange.

        It will be C{None} if we are not fully registered yet or if something
        bad happened during the last exchange and we could not get the token
        that the server had given us.
        """
        exchange_token = self._message_store.get_exchange_token()

        # Before starting the exchange set the saved token to None. This will
        # prevent us from locking ourselves out if the exchange fails or if we
        # crash badly, while the server has saved a new token that we couldn't
        # receive or persist (this works because if the token is None the
        # server will be forgiving and will authenticate us based only on the
        # secure ID we provide).
        self._message_store.set_exchange_token(None)
        self._message_store.commit()

        return exchange_token

    def _notify_impending_exchange(self):
        self._reactor.fire("impending-exchange")

    def _make_payload(self):
        """Return a dict representing the complete exchange payload.

        The payload will contain all pending messages eligible for
        delivery, up to a maximum of C{max_messages} as passed to
        the L{__init__} method.
        """
        store = self._message_store
        accepted_types_digest = self._hash_types(store.get_accepted_types())
        messages = store.get_pending_messages(self._max_messages)
        total_messages = store.count_pending_messages()
        if messages:
            # Each message is tagged with the API that the client was
            # using at the time the message got added to the store.  The
            # logic below will make sure that all messages which are added
            # to the payload being built will have the same api, and any
            # other messages will be postponed to the next exchange.
            server_api = messages[0].get("api")
            for i, message in enumerate(messages):
                if message.get("api") != server_api:
                    break
            else:
                i = None
            if i is not None:
                del messages[i:]

            # DEPRECATED Remove this once API 2.0 is gone:
            if server_api is None:
                # The per-message API logic was introduced on API 2.1, so a
                # missing API must be 2.0.
                server_api = "2.0"
        else:
            server_api = SERVER_API
        payload = {"server-api": server_api,
                   "client-api": CLIENT_API,
                   "sequence": store.get_sequence(),
                   "accepted-types": accepted_types_digest,
                   "messages": messages,
                   "total-messages": total_messages,
                   "next-expected-sequence": store.get_server_sequence()}
        accepted_client_types = self.get_client_accepted_message_types()
        accepted_client_types_hash = self._hash_types(accepted_client_types)
        if accepted_client_types_hash != self._client_accepted_types_hash:
            payload["client-accepted-types"] = accepted_client_types
        return payload

    def _hash_types(self, types):
        accepted_types_str = ";".join(types)
        return md5(accepted_types_str).digest()

    def _handle_result(self, payload, result):
        """Handle a response from the server.

        Called by L{exchange} after a batch of messages has been
        successfully delivered to the server.

        If the C{server_uuid} changed, a C{"server-uuid-changed"} event
        will be fired.

        Call L{handle_message} for each message in C{result}.

        @param payload: The payload that was sent to the server.
        @param result: The response got in reply to the C{payload}.
        """
        message_store = self._message_store
        self._client_accepted_types_hash = result.get(
            "client-accepted-types-hash")
        next_expected = result.get("next-expected-sequence")
        old_sequence = message_store.get_sequence()
        if next_expected is None:
            # If the server doesn't specify anything for the next-expected
            # value, just assume that it processed all messages that we sent
            # fine.
            next_expected = message_store.get_sequence()
            next_expected += len(payload["messages"])

        message_store_state = got_next_expected(message_store, next_expected)
        if message_store_state == ANCIENT:
            # The server has probably lost some data we sent it. The
            # slate has been wiped clean (by got_next_expected), now
            # let's fire an event to tell all the plugins that they
            # ought to generate new messages so the server gets some
            # up-to-date data.
            logging.info("Server asked for ancient data: resynchronizing all "
                         "state with the server.")
            self.send({"type": "resynchronize"})
            self._reactor.fire("resynchronize-clients")

        # Save the exchange token that the server has sent us. We will provide
        # it at the next exchange to prove that we're still the same client.
        # See also landscape.broker.transport.
        message_store.set_exchange_token(result.get("next-exchange-token"))

        old_uuid = message_store.get_server_uuid()
        new_uuid = result.get("server-uuid")
        if new_uuid != old_uuid:
            logging.info("Server UUID changed (old=%s, new=%s)."
                         % (old_uuid, new_uuid))
            self._reactor.fire("server-uuid-changed", old_uuid, new_uuid)
            message_store.set_server_uuid(new_uuid)

        message_store.commit()

        sequence = message_store.get_server_sequence()
        for message in result.get("messages", ()):
            self.handle_message(message)
            sequence += 1
            message_store.set_server_sequence(sequence)
            message_store.commit()

        if message_store.get_pending_messages(1):
            logging.info("Pending messages remain after the last exchange.")
            # Either the server asked us for old messages, or we
            # otherwise have more messages even after transferring
            # what we could.
            if next_expected != old_sequence:
                self.schedule_exchange(urgent=True)

    def register_message(self, type, handler):
        """Register a handler for the given message type.

        The C{handler} callable will to be executed when a message of
        type C{type} has been received from the server.

        Multiple handlers for the same type will be called in the
        order they were registered.
        """
        self._message_handlers.setdefault(type, []).append(handler)
        self._client_accepted_types.add(type)

    def handle_message(self, message):
        """
        Handle a message received from the server.

        Any message handlers registered with L{register_message} will
        be called.
        """
        if 'operation-id' in message:
            # This is a message that requires a response. Store the secure ID
            # so we can check for obsolete results later.
            self._exchange_store.add_message_context(
                message['operation-id'], self._registration_info.secure_id,
                message['type'])

        self._reactor.fire("message", message)
        # This has plan interference! but whatever.
        if message["type"] in self._message_handlers:
            for handler in self._message_handlers[message["type"]]:
                handler(message)

    def register_client_accepted_message_type(self, type):
        # stringify the type because it's a dbus.String.  It should work
        # anyway, but this is just for sanity and less confusing logs.
        self._client_accepted_types.add(str(type))

    def get_client_accepted_message_types(self):
        return sorted(self._client_accepted_types)


def get_accepted_types_diff(old_types, new_types):
    old_types = set(old_types)
    new_types = set(new_types)
    added_types = new_types - old_types
    stable_types = old_types & new_types
    removed_types = old_types - new_types
    diff = []
    diff.extend(["+%s" % type for type in added_types])
    diff.extend(["%s" % type for type in stable_types])
    diff.extend(["-%s" % type for type in removed_types])
    return " ".join(diff)



from landscape.tests.helpers import LandscapeTest, FakeBrokerServiceHelper
from landscape.reactor import FakeReactor
from landscape.monitor.config import MonitorConfiguration, ALL_PLUGINS
from landscape.monitor.service import MonitorService
from landscape.monitor.computerinfo import ComputerInfo
from landscape.monitor.loadaverage import LoadAverage


class MonitorServiceTest(LandscapeTest):

    helpers = [FakeBrokerServiceHelper]

    def setUp(self):
        super(MonitorServiceTest, self).setUp()
        config = MonitorConfiguration()
        config.load(["-c", self.config_filename])

        class FakeMonitorService(MonitorService):
            reactor_factory = FakeReactor

        self.service = FakeMonitorService(config)
        self.log_helper.ignore_errors("Typelib file for namespace")

    def test_plugins(self):
        """
        By default the L{MonitorService.plugins} list holds an instance of
        every enabled monitor plugin.
        """
        self.assertEqual(len(self.service.plugins), len(ALL_PLUGINS))

    def test_get_plugins(self):
        """
        If the C{--monitor-plugins} command line option is specified, only the
        given plugins will be enabled.
        """
        self.service.config.load(["--monitor-plugins",
                                  "ComputerInfo, LoadAverage"])
        plugins = self.service.get_plugins()
        self.assertTrue(isinstance(plugins[0], ComputerInfo))
        self.assertTrue(isinstance(plugins[1], LoadAverage))

    def test_start_service(self):
        """
        The L{MonitorService.startService} method connects to the broker,
        starts the plugins and register the monitor as broker client.  It also
        start listening on its own socket for incoming connections.
        """

        def stop_service(ignored):
            [connector] = self.broker_service.broker.get_connectors()
            connector.disconnect()
            self.service.stopService()
            self.broker_service.stopService()

        def assert_broker_connection(ignored):
            self.assertEqual(len(self.broker_service.broker.get_clients()), 1)
            self.assertIs(self.service.broker, self.service.monitor.broker)
            result = self.service.broker.ping()
            return result.addCallback(stop_service)

        self.broker_service.startService()
        started = self.service.startService()
        return started.addCallback(assert_broker_connection)

    def test_stop_service(self):
        """
        The L{MonitorService.stopService} method flushes the data before
        shutting down the monitor, and closes the connection with the broker.
        """
        self.service.monitor = self.mocker.mock()
        self.service.monitor.flush()
        self.service.connector = self.mocker.mock()
        self.service.connector.disconnect()
        self.service.publisher = self.mocker.mock()
        self.service.publisher.stop()
        self.mocker.replay()
        self.service.stopService()



import logging


class UpgraderConflict(Exception):
    """Two upgraders with the same version have been registered."""


class UpgradeManagerBase(object):
    """A simple upgrade system."""

    def __init__(self):
        self._upgraders = {}

    def register_upgrader(self, version, function):
        """
        @param version: The version number that this upgrader is
            upgrading the database to. This defines the order that
            upgraders are run.
        @param function: The function to call when applying
            upgraders. It must take a single object, the database that is
            being upgraded.
        """
        if version in self._upgraders:
            raise UpgraderConflict("%s is already registered as %s; "
                                   "not adding %s" % (
                    version, self._upgraders[version], function))
        self._upgraders[version] = function

    def get_version(self):
        """
        Get the 'current' version of any database that this
        UpgradeManager will be applied to.
        """
        keys = self._upgraders.keys()
        if keys:
            return max(keys)
        return 0

    def upgrader(self, version):
        """
        A decorator for specifying that a function is an upgrader for
        this upgrade manager.

        @param version: The version number that the function will be
            upgrading to.
        """
        def inner(function):
            self.register_upgrader(version, function)
            return function
        return inner


class UpgradeManager(UpgradeManagerBase):

    def apply(self, persist):
        """Bring the database up-to-date.

        @param persist: The database to upgrade. It will be passed to
            all upgrade functions.
        """
        if not persist.has("system-version"):
            persist.set("system-version", 0)
        for version, upgrader in sorted(self._upgraders.items()):
            if version > persist.get("system-version"):
                persist.set("system-version", version)
                upgrader(persist)
                logging.info("Successfully applied patch %s" % version)

    def initialize(self, persist):
        """
        Mark the database as being up-to-date; use this when
        initializing a new database.
        """
        persist.set("system-version", self.get_version())


class SQLiteUpgradeManager(UpgradeManagerBase):
    """An upgrade manager backed by sqlite."""

    def get_database_versions(self, cursor):
        cursor.execute("SELECT version FROM patch")
        result = cursor.fetchall()
        return set([row[0] for row in result])

    def get_database_version(self, cursor):
        cursor.execute("SELECT MAX(version) FROM patch")
        version = cursor.fetchone()[0]
        if version:
           return version
        return 0

    def apply(self, cursor):
        """Bring the database up-to-date."""
        versions = self.get_database_versions(cursor)
        for version, upgrader in sorted(self._upgraders.items()):
            if version not in versions:
                self.apply_one(version, cursor)

    def apply_one(self, version, cursor):
        upgrader = self._upgraders[version]
        upgrader(cursor)
        cursor.execute("INSERT INTO patch VALUES (?)", (version,))

    def initialize(self, cursor):
        """
        Mark the database as being up-to-date; use this when
        initializing a new SQLite database.
        """
        cursor.execute("CREATE TABLE patch (version INTEGER)")
        for version, upgrader in sorted(self._upgraders.items()):
            cursor.execute("INSERT INTO patch VALUES (?)", (version,))



import logging
import os
import re

from landscape.plugin import PluginConfigError
from landscape.monitor.plugin import MonitorPlugin


class ProcessorInfo(MonitorPlugin):
    """Plugin captures information about the processor(s) in this machine.

    This plugin runs once per client session.  When processor
    information is retrieved it's compared against the last known
    processor information, which is saved in persistent storage.  A
    message is only put on the message queue if the latest processor
    information differs from the last known processor information.

    The information available from /proc/cpuinfo varies per platform.
    For example, an Apple PowerMac Dual G5 doesn't contain a vendor ID
    and provides the processor name in the 'cpu' field, as opposed to
    the 'model name' field used on x86-based hardware.  For reasons
    such as this, the schema of the data reported by this plugin is
    flexible.  Only 'processor-id' and 'model' are guaranteed to be
    present.

    In order to deal with the vagaries of parsing /proc/cpu
    information on the various platforms we support, message
    generation is deferred to per-platform message factories.

    @param delay: Set the starting delay.
    @param machine_name: The machine name to report.
    @param source_filename: The filesystem path to read information from.
    """

    persist_name = "processor-info"
    scope = "cpu"
    # Prevent the Plugin base-class from scheduling looping calls.
    run_interval = None

    def __init__(self, delay=2, machine_name=None,
                 source_filename="/proc/cpuinfo"):
        self._delay = delay
        self._source_filename = source_filename

        if machine_name is None:
            machine_name = os.uname()[4]

        self._cpu_info_reader = self._create_cpu_info_reader(machine_name,
                                                             source_filename)

    def _create_cpu_info_reader(self, machine_name, source_filename):
        """Return a message factory suitable for the specified machine name."""
        for pair in message_factories:
            regexp = re.compile(pair[0])

            if regexp.match(machine_name):
                return pair[1](source_filename)

        raise PluginConfigError("A processor info reader for '%s' is not "
                                "available." % machine_name)

    def register(self, registry):
        """Register this plugin with the specified plugin registry."""
        super(ProcessorInfo, self).register(registry)
        self.registry.reactor.call_later(self._delay, self.run)
        self.call_on_accepted("processor-info", self.send_message, True)

    def create_message(self):
        """Retrieve processor information and generate a message."""
        return {"type": "processor-info",
                "processors": self._cpu_info_reader.create_message()}

    def send_message(self, urgent=False):
        dirty = False
        message = self.create_message()

        for processor in message["processors"]:
            key = ("processor", str(processor["processor-id"]))
            cached_processor = self._persist.get(key)
            if cached_processor is None:
                cached_processor = {}
                self._update(cached_processor, processor)
                dirty = True
            else:
                if self._has_changed(cached_processor, processor):
                    self._update(cached_processor, processor)
                    dirty = True

        if dirty:
            logging.info("Queueing message with updated processor info.")
            self.registry.broker.send_message(
                message, self._session_id, urgent=urgent)

    def run(self, urgent=False):
        """Create a message and put it on the message queue."""
        self.registry.broker.call_if_accepted("processor-info",
                                              self.send_message, urgent)

    def _has_changed(self, processor, message):
        """Returns true if processor details changed since the last read."""
        if processor["model"] != message["model"]:
            return True

        if processor["vendor"] != message.get("vendor", ""):
            return True

        if processor["cache_size"] != message.get("cache-size", -1):
            return True

        return False

    def _update(self, processor, message):
        """Update the processor details with current values."""
        processor["id"] = message["processor-id"]
        processor["model"] = message["model"]
        processor["cache_size"] = message.get("cache-size", -1)
        processor["vendor"] = message.get("vendor", "")
        self._persist.set(("processor", str(message["processor-id"])),
                          processor)


class PowerPCMessageFactory:
    """Factory for ppc-based processors provides processor information.

    @param source_filename: The file name of the data source.
    """

    def __init__(self, source_filename):
        self._source_filename = source_filename

    def create_message(self):
        """Returns a list containing information about each processor."""
        processors = []
        file = open(self._source_filename)

        try:
            current = None

            for line in file:
                parts = line.split(":", 1)
                key = parts[0].strip()

                if key == "processor":
                    current = {"processor-id": int(parts[1].strip())}
                    processors.append(current)
                elif key == "cpu":
                    current["model"] = parts[1].strip()
        finally:
            file.close()

        return processors


class ARMMessageFactory:
    """Factory for arm-based processors provides processor information.

    @param source_filename: The file name of the data source.
    """

    def __init__(self, source_filename):
        self._source_filename = source_filename

    def create_message(self):
        """Returns a list containing information about each processor."""
        processors = []
        file = open(self._source_filename)

        try:
            regexp = re.compile("(?P<key>.*?)\s*:\s*(?P<value>.*)")
            current = {}

            for line in file:
                match = regexp.match(line.strip())
                if match:
                    key = match.group("key")
                    value = match.group("value")

                    if key == "Processor":
                        # ARM doesn't support SMP, thus no processor-id in
                        # the cpuinfo
                        current["processor-id"] = 0
                        current["model"] = value
                    elif key == "Cache size":
                        current["cache-size"] = int(value)

            if current:
                processors.append(current)
        finally:
            file.close()

        return processors


class SparcMessageFactory:
    """Factory for sparc-based processors provides processor information.

    @param source_filename: The file name of the data source.
    """

    def __init__(self, source_filename):
        self._source_filename = source_filename

    def create_message(self):
        """Returns a list containing information about each processor."""
        processors = []
        model = None
        file = open(self._source_filename)

        try:
            regexp = re.compile("CPU(\d{1})+")

            for line in file:
                parts = line.split(":", 1)
                key = parts[0].strip()

                if key == "cpu":
                    model = parts[1].strip()
                elif regexp.match(key):
                    start, end = re.compile("\d+").search(key).span()
                    message = {"processor-id": int(key[start:end]),
                               "model": model}
                    processors.append(message)
        finally:
            file.close()

        return processors


class X86MessageFactory:
    """Factory for x86-based processors provides processor information.

    @param source_filename: The file name of the data source.
    """

    def __init__(self, source_filename):
        self._source_filename = source_filename

    def create_message(self):
        """Returns a list containing information about each processor."""
        processors = []
        file = open(self._source_filename)

        try:
            current = None

            for line in file:
                parts = line.split(":", 1)
                key = parts[0].strip()

                if key == "processor":
                    current = {"processor-id": int(parts[1].strip())}
                    processors.append(current)
                elif key == "vendor_id":
                    current["vendor"] = parts[1].strip()
                elif key == "model name":
                    current["model"] = parts[1].strip()
                elif key == "cache size":
                    value_parts = parts[1].split()
                    current["cache-size"] = int(value_parts[0].strip())
        finally:
            file.close()

        return processors


message_factories = [("arm*", ARMMessageFactory),
                     ("ppc(64)?", PowerPCMessageFactory),
                     ("sparc[64]", SparcMessageFactory),
                     ("i[3-7]86|x86_64", X86MessageFactory)]



# Copyright (c) 2001-2007 Twisted Matrix Laboratories.

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Copies of certain classes from Twisted 2.5, so that we can use the
functionality they provide with Twisted 2.2 and up. These should
really only be used from the test suite for now.

Currently:

 * L{twisted.internet.task.Clock}, which is new in Twisted 2.5.
 * L{twisted.internet.base.DelayedCall}, which didn't grow its
   C{seconds} argument until after Twisted 2.2.
"""


from twisted.internet import error
from twisted.python.runtime import seconds as runtimeSeconds
from twisted.python import reflect
import traceback


class Clock:
    """
    Provide a deterministic, easily-controlled implementation of
    L{IReactorTime.callLater}.  This is commonly useful for writing
    deterministic unit tests for code which schedules events using this API.
    """
    rightNow = 0.0

    def __init__(self):
        self.calls = []

    def seconds(self):
        """
        Pretend to be time.time().  This is used internally when an operation
        such as L{IDelayedCall.reset} needs to determine a a time value
        relative to the current time.

        @rtype: C{float}
        @return: The time which should be considered the current time.
        """
        return self.rightNow


    def callLater(self, when, what, *a, **kw):
        """
        See L{twisted.internet.interfaces.IReactorTime.callLater}.
        """
        self.calls.append(
            DelayedCall(self.seconds() + when,
                             what, a, kw,
                             self.calls.remove,
                             lambda c: None,
                             self.seconds))
        self.calls.sort(lambda a, b: cmp(a.getTime(), b.getTime()))
        return self.calls[-1]


    def advance(self, amount):
        """
        Move time on this clock forward by the given amount and run whatever
        pending calls should be run.

        @type amount: C{float}
        @param amount: The number of seconds which to advance this clock's
        time.
        """
        self.rightNow += amount
        while self.calls and self.calls[0].getTime() <= self.seconds():
            call = self.calls.pop(0)
            call.called = 1
            call.func(*call.args, **call.kw)


    def pump(self, timings):
        """
        Advance incrementally by the given set of times.

        @type timings: iterable of C{float}
        """
        for amount in timings:
            self.advance(amount)



class DelayedCall:

    # enable .debug to record creator call stack, and it will be logged if
    # an exception occurs while the function is being run
    debug = False
    _str = None

    def __init__(self, time, func, args, kw, cancel, reset,
                 seconds=runtimeSeconds):
        """
        @param time: Seconds from the epoch at which to call C{func}.
        @param func: The callable to call.
        @param args: The positional arguments to pass to the callable.
        @param kw: The keyword arguments to pass to the callable.
        @param cancel: A callable which will be called with this
            DelayedCall before cancellation.
        @param reset: A callable which will be called with this
            DelayedCall after changing this DelayedCall's scheduled
            execution time. The callable should adjust any necessary
            scheduling details to ensure this DelayedCall is invoked
            at the new appropriate time.
        @param seconds: If provided, a no-argument callable which will be
            used to determine the current time any time that information is
            needed.
        """
        self.time, self.func, self.args, self.kw = time, func, args, kw
        self.resetter = reset
        self.canceller = cancel
        self.seconds = seconds
        self.cancelled = self.called = 0
        self.delayed_time = 0
        if self.debug:
            self.creator = traceback.format_stack()[:-2]

    def getTime(self):
        """Return the time at which this call will fire

        @rtype: C{float}
        @return: The number of seconds after the epoch at which this call is
        scheduled to be made.
        """
        return self.time + self.delayed_time

    def cancel(self):
        """Unschedule this call

        @raise AlreadyCancelled: Raised if this call has already been
        unscheduled.

        @raise AlreadyCalled: Raised if this call has already been made.
        """
        if self.cancelled:
            raise error.AlreadyCancelled
        elif self.called:
            raise error.AlreadyCalled
        else:
            self.canceller(self)
            self.cancelled = 1
            if self.debug:
                self._str = str(self)
            del self.func, self.args, self.kw

    def reset(self, secondsFromNow):
        """Reschedule this call for a different time

        @type secondsFromNow: C{float}
        @param secondsFromNow: The number of seconds from the time of the
        C{reset} call at which this call will be scheduled.

        @raise AlreadyCancelled: Raised if this call has been cancelled.
        @raise AlreadyCalled: Raised if this call has already been made.
        """
        if self.cancelled:
            raise error.AlreadyCancelled
        elif self.called:
            raise error.AlreadyCalled
        else:
            newTime = self.seconds() + secondsFromNow
            if newTime < self.time:
                self.delayed_time = 0
                self.time = newTime
                self.resetter(self)
            else:
                self.delayed_time = newTime - self.time

    def delay(self, secondsLater):
        """Reschedule this call for a later time

        @type secondsLater: C{float}
        @param secondsLater: The number of seconds after the originally
        scheduled time for which to reschedule this call.

        @raise AlreadyCancelled: Raised if this call has been cancelled.
        @raise AlreadyCalled: Raised if this call has already been made.
        """
        if self.cancelled:
            raise error.AlreadyCancelled
        elif self.called:
            raise error.AlreadyCalled
        else:
            self.delayed_time += secondsLater
            if self.delayed_time < 0:
                self.activate_delay()
                self.resetter(self)

    def activate_delay(self):
        self.time += self.delayed_time
        self.delayed_time = 0

    def active(self):
        """Determine whether this call is still pending

        @rtype: C{bool}
        @return: True if this call has not yet been made or cancelled,
        False otherwise.
        """
        return not (self.cancelled or self.called)

    def __le__(self, other):
        return self.time <= other.time

    def __str__(self):
        if self._str is not None:
            return self._str
        if hasattr(self, 'func'):
            if hasattr(self.func, 'func_name'):
                func = self.func.func_name
                if hasattr(self.func, 'im_class'):
                    func = self.func.im_class.__name__ + '.' + func
            else:
                func = reflect.safe_repr(self.func)
        else:
            func = None

        now = self.seconds()
        L = ["<DelayedCall %s [%ss] called=%s cancelled=%s" % (
                id(self), self.time - now, self.called, self.cancelled)]
        if func is not None:
            L.extend((" ", func, "("))
            if self.args:
                L.append(", ".join([reflect.safe_repr(e) for e in self.args]))
                if self.kw:
                    L.append(", ")
            if self.kw:
                L.append(", ".join(['%s=%s' % (k, reflect.safe_repr(v)) for (k, v) in self.kw.iteritems()]))
            L.append(")")

        if self.debug:
            L.append("\n\ntraceback at creation: \n\n%s" % ('    '.join(self.creator)))
        L.append('>')

        return "".join(L)



from twisted.internet.defer import Deferred

from landscape.tests.helpers import LandscapeTest
from landscape.tests.helpers import ManagerHelper
from landscape.manager.plugin import ManagerPlugin, SUCCEEDED, FAILED


class BrokerPluginTest(LandscapeTest):

    helpers = [ManagerHelper]

    def test_call_with_operation_result_success(self):
        """
        A helper method exists which calls a function and sends an
        operation-result message based on the success of that method.
        """
        plugin = ManagerPlugin()
        plugin.register(self.manager)
        broker_service = self.broker_service
        broker_service.message_store.set_accepted_types(["operation-result"])
        message = {"operation-id": 12312}
        operation = lambda: None

        def assert_messages(ignored):
            messages = broker_service.message_store.get_pending_messages()
            self.assertMessages(messages,
                                [{"type": "operation-result",
                                  "status": SUCCEEDED,
                                  "operation-id": 12312}])

        result = plugin.call_with_operation_result(message, operation)
        return result.addCallback(assert_messages)

    def test_call_with_operation_result_error(self):
        """
        The helper for operation-results sends an appropriate message when an
        exception is raised from the given method.
        """
        self.log_helper.ignore_errors(RuntimeError)
        plugin = ManagerPlugin()
        plugin.register(self.manager)
        broker_service = self.broker_service
        broker_service.message_store.set_accepted_types(["operation-result"])
        message = {"operation-id": 12312}

        def operation():
            raise RuntimeError("What the crap!")

        def assert_messages(ignored):
            messages = broker_service.message_store.get_pending_messages()
            self.assertMessages(messages,
                                [{"type": "operation-result", "status": FAILED,
                                  "result-text": "RuntimeError: What the "
                                  "crap!", "operation-id": 12312}])
            logdata = self.logfile.getvalue()
            self.assertTrue("RuntimeError: What the crap!" in logdata, logdata)

        result = plugin.call_with_operation_result(message, operation)
        return result.addCallback(assert_messages)

    def test_call_with_operation_result_exchanges_urgently(self):
        """
        Operation results are reported to the server as quickly as possible.
        """
        plugin = ManagerPlugin()
        plugin.register(self.manager)
        broker_service = self.broker_service
        broker_service.message_store.set_accepted_types(["operation-result"])
        message = {"operation-id": 123}
        operation = lambda: None

        def assert_urgency(ignored):
            self.assertTrue(broker_service.exchanger.is_urgent())

        result = plugin.call_with_operation_result(message, operation)
        return result.addCallback(assert_urgency)

    def test_callable_returning_a_deferred(self):
        """
        The callable parameter can return a C{Deferred}.
        """
        plugin = ManagerPlugin()
        plugin.register(self.manager)
        broker_service = self.broker_service
        broker_service.message_store.set_accepted_types(["operation-result"])
        message = {"operation-id": 12312}
        deferred = Deferred()
        operation = lambda: deferred

        def assert_messages(ignored):
            messages = broker_service.message_store.get_pending_messages()
            self.assertMessages(messages,
                                [{"type": "operation-result",
                                  "result-text": "blah",
                                  "status": SUCCEEDED,
                                  "operation-id": 12312}])

        result = plugin.call_with_operation_result(message, operation)
        result.addCallback(assert_messages)
        deferred.callback("blah")
        return result
